# -*- coding: utf-8 -*-

"""
MIT License

Copyright (c) 2019 Terbau

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import datetime

replacers = {
    'placetop1': 'wins'
}




class StatsV2:
    """Represents a users Battle Royale stats on Fortnite.
    
    Attributes
    ----------
    stats: :class:`dict`
        :class:`dict` including stats.
    start_time: :class:`datetime.datetime`
        The UTC start time of the stats retrieved.
    end_time: :class:`datetime`
        The UTC end time of the stats retrieved. 
    """
    def __init__(self, data):
        self.raw = data

        self.stats = {}
        # self.start_time = datetime.datetime.fromtimestamp(data['startTime'] / 1000)
        # self.end_time = datetime.datetime.fromtimestamp(data['endTime'] / 1000)

        self._parse()

    @staticmethod
    def create_stat(stat, platform, playlist):
        if stat in replacers.values():
            for k, v in replacers.items():
                if v == stat:
                    stat = k

        return 'br_{0}_{1}_m0_playlist_{2}'.format(stat, platform.value, playlist)

    def get_kd(self, data):
        """Gets the kd of a gamemode
        
        Usage: ::
            
            # gets ninjas kd in solo on input touch
            async def get_ninja_touch_solo_kd():
                profile = await client.fetch_profile('Ninja)
                stats = await client.fetch_br_stats(profile.id)

                return stats.get_kd(stats.stats['touch']['defaultsolo'])
        
        Parameters
        ----------
        data: :class:`dict`
            A :class:`dict` which includes the keys: ``kills``, ``matchesplayed`` and ``wins``.

        Returns
        -------
        :class:`float`
            Returns the kd with a decimal point accuracy of two.
        """

        kills = data.get('kills', 0)
        matches = data.get('matchesplayed', 0)
        wins = data.get('wins', 0)

        try:
            kd = kills / (matches - wins)
        except ZeroDivisionError:
            kd = 0
        return float(format(kd, '.2f'))
    
    def get_winpercentage(self, data):
        """Gets the winpercentage of a gamemode
        
        Usage: ::
            
            # gets ninjas winpercentage in solo on input touch
            async def get_ninja_touch_solo_winpercentage():
                profile = await client.fetch_profile('Ninja)
                stats = await client.fetch_br_stats(profile.id)

                return stats.get_winpercentage(stats.stats['touch']['defaultsolo'])

        Parameters
        ----------
        data: :class:`dict`
            A :class:`dict` which includes the keys: matchesplayed`` and ``wins``.

        Returns
        -------
        :class:`float`
            Returns the winpercentage with a decimal point accuracy of two.
        """

        matches = data.get('matchesplayed', 0)
        wins = data.get('wins', 0)

        try:
            winper = (wins * 100) / matches
        except ZeroDivisionError:
            winper = 0
        if winper > 100:
            winper = 100
        return float(format(winper, '.2f'))

    def _parse(self):
        result = {}
        for fullname, stat in self.raw['stats'].items():
            parts = fullname.split('_')

            name = parts[1]
            inp = parts[2]
            playlist = '_'.join(parts[5:])

            try:
                name = replacers[name]
            except KeyError:
                pass

            if name == 'lastmodified':
                stat = datetime.datetime.fromtimestamp(stat)
            
            if inp not in result.keys():
                result[inp] = {}
            if playlist not in result[inp].keys():
                result[inp][playlist] = {}
            
            result[inp][playlist][name] = stat
        self.stats = result




# -*- coding: utf-8 -*-

"""
MIT License

Copyright (c) 2019 Terbau

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import datetime
import json
import asyncio
import random
import weakref
import aioxmpp
import re

from .errors import FortniteException, PartyError, PartyPermissionError, HTTPException
from .user import User
from .friend import Friend
from .enums import PartyPrivacy, DefaultCharacters


class MetaBase:
    def __init__(self):
        self.schema = {}

    def set_prop(self, prop, value, raw=False):
        if raw:
            self.schema[prop] = str(value)
            return self.schema[prop]
        
        _t = prop[-1:]
        if _t == 'j':
            self.schema[prop] = json.dumps(value)
        elif _t == 'U':
            self.schema[prop] = int(value)
        else:
            self.schema[prop] = str(value)
        return self.schema[prop]
    
    def get_prop(self, prop, raw=False):
        if raw:
            self.schema.get(prop)
        
        _t = prop[-1:]
        _v = self.schema.get(prop)
        if _t == 'b':
            return False if _v is None or (isinstance(_v, str) and _v.lower() == 'false') else True
        elif _t == 'j':
            return {} if _v is None else json.loads(_v)
        elif _t == 'U':
            return {} if _v is None else int(_v)
        else:
            return '' if _v is None else str(_v)

    def update(self, schema, raw=False):
        if schema is None: 
            return

        for prop, value in schema.items():
            self.set_prop(prop, value, raw=raw)
    
    def remove(self, schema):
        for prop in schema:
            try:
                del self.schema[prop]
            except KeyError:
                pass


class PartyMemberMeta(MetaBase):
    def __init__(self, member, meta=None):
        super().__init__()
        self.member = member

        character = (random.choice(list(DefaultCharacters))).name
        self.schema = {
            'Location_s': 'PreLobby',
            'CampaignHero_j': json.dumps({
                'CampaignHero': {
                    'heroItemInstanceId': '',
                    'heroType': "FortHeroType'/Game/Athena/Heroes/{0}.{0}'" \
                                "".format(character),
                },
            }),
            'MatchmakingLevel_U': '0',
            'ZoneInstanceId_s': '',
            'HomeBaseVersion_U': '1',
            'HasPreloadedAthena_b': False,
            'FrontendEmote_j': json.dumps({
                'FrontendEmote': {
                    'emoteItemDef': 'None',
                    'emoteItemDefEncryptionKey': '',
                    'emoteSection': -1,
                },
            }),
            'NumAthenaPlayersLeft_U': '0',
            'UtcTimeStartedMatchAthena_s': '0001-01-01T00:00:00.000Z',
            'GameReadiness_s': 'NotReady',
            'HiddenMatchmakingDelayMax_U': '0',
            'ReadyInputType_s': 'Count',
            'CurrentInputType_s': 'MouseAndKeyboard',
            'AssistedChallengeInfo_j': json.dumps({
                'AssistedChallengeInfo': {
                    'questItemDef': 'None',
                    'objectivesCompleted': 0,
                },
            }),
            'MemberSquadAssignmentRequest_j': json.dumps({
                'MemberSquadAssignmentRequest': {
                    'startingAbsoluteIdx': -1,
                    'targetAbsoluteIdx': -1,
                    'swapTargetMemberId': 'INVALID',
                    'version': 0,
                },
            }),
            'AthenaCosmeticLoadout_j': json.dumps({
                'AthenaCosmeticLoadout': {
                    'characterDef': "AthenaCharacterItemDefinition'/Game/Athena/Items/Cosmetics/Characters/{0}.{0}'" \
                                    "".format(character),
                    'characterEKey': '',
                    'backpackDef': 'None',
                    'backpackEKey': '',
                    'pickaxeDef': "AthenaPickaxeItemDefinition'/Game/Athena/Items/Cosmetics/" \
                                  "Pickaxes/DefaultPickaxe.DefaultPickaxe'",
                    'pickaxeEKey': '',
                    'variants': [],
                },
            }),
            'AthenaBannerInfo_j': json.dumps({
                'AthenaBannerInfo': {
                    'bannerIconId': 'standardbanner15',
                    'bannerColorId': 'defaultcolor15',
                    'seasonLevel': 1,
                },
            }),
            'BattlePassInfo_j': json.dumps({
                'BattlePassInfo': {
                    'bHasPurchasedPass': False,
                    'passLevel': 1,
                    'selfBoostXp': 0,
                    'friendBoostXp': 0,
                },
            }),
            'Platform_j': json.dumps({
                'Platform': {
                    'platformStr': self.member.client.platform,
                },
            }),
            'PlatformUniqueId_s': 'INVALID',
            'PlatformSessionId_s': '',
            'CrossplayPreference_s': 'OptedIn',
            'VoiceChatEnabled_b': 'true',
            'VoiceConnectionId_s': '',
        }

        if meta is not None:
            self.update(meta, raw=True)

    @property
    def ready(self):
        base = self.get_prop('GameReadiness_s')
        return True if base == "Ready" else False

    @property
    def input(self):
        return self.get_prop('CurrentInputType_s')

    @property
    def assisted_challenge(self):
        base = self.get_prop('AssistedChallengeInfo_j')
        result = result = re.search(r".*\.(.*)", base['AssistedChallengeInfo']['questItemDef'].strip("'"))

        if result is None or result[1] == 'None':
            return None
        return result[1]

    @property
    def outfit(self):
        base = self.get_prop('AthenaCosmeticLoadout_j')
        return re.search(r'.*(CID.*)\..*', base['AthenaCosmeticLoadout']['characterDef'])[1]

    @property
    def outfit_variants(self):
        base = self.get_prop('AthenaCosmeticLoadout_j')
        return base['AthenaCosmeticLoadout']['variants']

    @property
    def backpack(self):
        base = self.get_prop('AthenaCosmeticLoadout_j')
        result = re.search(r".*\.(.*)", base['AthenaCosmeticLoadout']['backpackDef'].strip("'"))

        if result is None or result[1] == 'None':
            return None
        return result[1]

    @property
    def pickaxe(self):
        base = self.get_prop('AthenaCosmeticLoadout_j')
        result = re.search(r".*\.(.*)", base['AthenaCosmeticLoadout']['pickaxeDef'].strip("'"))

        if result is None or result[1] == 'None':
            return None
        return result[1]

    @property
    def emote(self):
        base = self.get_prop('FrontendEmote_j')
        result = re.search(r'.*(EID.*)\..*', base['FrontendEmote']['emoteItemDef'])

        if result is None or result[1] == 'None':
            return None
        return result[1]

    @property
    def banner(self):
        base = self.get_prop('AthenaBannerInfo_j')
        banner_info = base['AthenaBannerInfo']

        return (banner_info['bannerIconId'], banner_info['bannerColorId'], banner_info['seasonLevel'])

    @property
    def battlepass_info(self):
        base = self.get_prop('BattlePassInfo_j')
        bp_info = base['BattlePassInfo']

        return (bp_info['bHasPurchasedPass'], bp_info['passLevel'], bp_info['selfBoostXp'], bp_info['friendBoostXp'])
        
    @property
    def platform(self):
        base = self.get_prop('Platform_j')
        return base['Platform']['platformStr']

    def set_readiness(self, val):
        return {'GameReadiness_s': self.set_prop('GameReadiness_s', val)}

    def set_emote(self, emote=None, emote_ekey=None, section=None):
        data = (self.get_prop('FrontendEmote_j'))['FrontendEmote']
        
        if emote:
            data['emoteItemDef'] = emote
        if emote_ekey:
            data['emoteItemDefEncryptionKey'] = emote_ekey
        if section:
            data['emoteSection'] = section
        
        final = {'FrontendEmote': data}
        return {'FrontendEmote_j': self.set_prop('FrontendEmote_j', final)}

    def set_assisted_challenge(self, quest=None, completed=None):
        data = (self.get_prop('AssistedChallengeInfo_j'))['AssistedChallenge_j']

        if quest:
            data['questItemDef'] = quest
        if completed:
            data['objectivesCompleted'] = completed
        
        final = {'AssistedChallengeInfo': data}
        return {'AssistedChallengeInfo_j': self.set_prop('AssistedChallengeInfo_j', final)}

    def set_banner(self, banner_icon=None, banner_color=None, season_level=None):
        data = (self.get_prop('AthenaBannerInfo_j'))['AthenaBannerInfo']

        if banner_icon:
            data['bannerIconId'] = banner_icon
        if banner_color:
            data['bannerColorId'] = banner_color
        if season_level:
            data['seasonLevel'] = season_level
        
        final = {'AthenaBannerInfo': data}
        return {'AthenaBannerInfo_j': self.set_prop('AthenaBannerInfo_j', final)}

    def set_battlepass_info(self, has_purchased=None, level=None, self_boost_xp=None, 
                            friend_boost_xp=None):
        data = (self.get_prop('BattlePassInfo_j'))['BattlePassInfo']

        if has_purchased:
            data['bHasPurchasedPass'] = has_purchased
        if level:
            data['passLevel'] = level
        if self_boost_xp:
            data['selfBoostXp'] = self_boost_xp
        if friend_boost_xp:
            data['friendBoostXp'] = friend_boost_xp
        
        final = {'BattlePassInfo': data}
        return {'BattlePassInfo_j': self.set_prop('BattlePassInfo_j', final)}

    def set_cosmetic_loadout(self, character=None, character_ekey=None, backpack=None,
                             backpack_ekey=None, pickaxe=None, pickaxe_ekey=None, variants=None):
        data = (self.get_prop('AthenaCosmeticLoadout_j'))['AthenaCosmeticLoadout']

        if character:
            data['characterDef'] = character
        if character_ekey:
            data['characterEKey'] = character_ekey
        if backpack:
            data['backpackDef'] = backpack
        if backpack_ekey:
            data['backpackEKey'] = backpack_ekey
        if pickaxe:
            data['pickaxeDef'] = pickaxe
        if pickaxe_ekey:
            data['pickaxeEKey'] = pickaxe_ekey
        if variants:
            data['variants'] = variants
        
        final = {'AthenaCosmeticLoadout': data}
        return {'AthenaCosmeticLoadout_j': self.set_prop('AthenaCosmeticLoadout_j', final)}


class PartyMeta(MetaBase):
    def __init__(self, party, meta=None):
        super().__init__()
        self.party = party

        self.schema = {
            'PrimaryGameSessionId_s': '',
            'PartyState_s': 'BattleRoyaleView',
            'LobbyConnectionStarted_b': 'false',
            'MatchmakingResult_s': 'NoResults',
            'MatchmakingState_s': 'NotMatchmaking',
            'SessionIsCriticalMission_b': 'false',
            'ZoneTileIndex_U': '-1',
            'ZoneInstanceId_s': '',
            'TheaterId_s': '',
            'TileStates_j': json.dumps({
                'TileStates': [],
            }),
            'MatchmakingInfoString_s': '',
            'CustomMatchKey_s': '',
            'PlaylistData_j': json.dumps({
                'PlaylistData': {
                    'playlistName': 'Playlist_DefaultDuo',
                    'tournamentId': '',
                    'eventWindowId': '',
                    'regionId': 'EU',
                },
            }),
            'AthenaSquadFill_b': 'true',
            'AllowJoinInProgress_b': 'false',
            'LFGTime_s': '0001-01-01T00:00:00.000Z',
            'PartyIsJoinedInProgress_b': 'false',
            'GameSessionKey_s': '',
            'RawSquadAssignments_j': '',
            'PrivacySettings_j': json.dumps({
                'PrivacySettings': {
                    'partyType': self.party.config['privacy']['partyType'],
                    'partyInviteRestriction': self.party.config['privacy']['inviteRestriction'],
                    'bOnlyLeaderFriendsCanJoin': self.party.config['privacy']['onlyLeaderFriendsCanJoin'],
                },
            }),
            'PlatformSessions_j': json.dumps({
                'PlatformSessions': [],
            }),
        }

        if meta is not None:
            self.update(meta, raw=True)
    
    @property
    def playlist_info(self):
        base = self.get_prop('PlaylistData_j')
        info = base['PlaylistData']

        return (info['playlistName'], info['tournamentId'], info['eventWindowId'], info['regionId'])

    @property
    def squad_fill_enabled(self):
        return self.get_prop('AthenaSquadFill_b')

    @property
    def privacy(self):
        curr_priv = (self.get_prop('PrivacySettings_j'))['PrivacySettings']

        for privacy in PartyPrivacy:
            if curr_priv['partyType'] != privacy.value['partyType']:
                continue
            
            try:
                if curr_priv['partyInviteRestriction'] != privacy.value['partyInviteRestriction']:
                    continue

                if curr_priv['bOnlyLeaderFriendsCanJoin'] != privacy.value['bOnlyLeaderFriendsCanJoin']:
                    continue
            except KeyError:
                pass

            return privacy

    def set_playlist(self, playlist=None, tournament=None, event_window=None, region=None):
        data = (self.get_prop('PlaylistData_j'))['PlaylistData']

        if playlist:
            data['playlistName'] = playlist
        if tournament:
            data['tournamentId'] = tournament
        if event_window:
            data['eventWindowId'] = event_window
        if region:
            data['regionId'] = region
        
        final = {'PlaylistData': data}
        return {'PlaylistData_j': self.set_prop('PlaylistData_j', final)}

    def set_custom_key(self, key):
        return {'CustomMatchKey_s': self.set_prop('CustomMatchKey_s', key)}

    def set_fill(self, val):
        return {'AthenaSquadFill_b': self.set_prop('AthenaSquadFill_b', (str(val)).lower())}

    def set_privacy(self, privacy):
        updated = {}
        deleted = []

        p = self.get_prop('PrivacySettings_j')
        if p:
            updated['PrivacySettings_j'] = self.set_prop('PrivacySettings_j', {
                'PrivacySettings': {
                    **p['PrivacySettings'],
                    'partyType': privacy['partyType'],
                    'bOnlyLeaderFriendsCanJoin': privacy['onlyLeaderFriendsCanJoin'],
                    'partyInviteRestriction': privacy['inviteRestriction'],
                }
            })
        
        updated['urn:epic:cfg:presence-perm_s'] = self.set_prop(
            'urn:epic:cfg:presence-perm_s', 
            privacy['presencePermission'],
        )

        updated['urn:epic:cfg:accepting-members_b'] = self.set_prop(
            'urn:epic:cfg:accepting-members_b',
            str(privacy['acceptingMembers']).lower(),
        )

        updated['urn:epic:cfg:invite-perm_s'] = self.set_prop(
            'urn:epic:cfg:invite-perm_s',
            privacy['invitePermission'],
        )

        if privacy['partyType'] not in ('Public', 'FriendsOnly'):
            deleted.append('urn:epic:cfg:not-accepting-members')

        if privacy['partyType'] == 'Private':
            updated['urn:epic:cfg:not-accepting-members-reason_i'] = 7
        else:
            deleted.append('urn:epic:cfg:not-accepting-members-reason_i')
        
        return (updated, deleted)
    
    def refresh_squad_assignments(self):
        assignments = []

        i = 0
        for member in self.party.members.values():
            if member.is_leader:
                assignments.append({
                    'memberId': member.id,
                    'absoluteMemberIdx': 0,
                })
            else:
                i += 1
                assignments.append({
                    'memberId': member.id,
                    'absoluteMemberIdx': i,
                })
        
        return self.set_prop('RawSquadAssignments_j', {
            'RawSquadAssignments': assignments
        })   


class PartyMemberBase(User):
    def __init__(self, client, party, data):
        super().__init__(client=client, data=data)

        self._party = party

        self._joined_at = self.client.from_iso(data['joined_at'])
        self.meta = PartyMemberMeta(self, meta=data.get('meta'))
        self._update(data)

    @property
    def party(self):
        """Union[:class:`Party`, :class:`ClientParty`]: The party this member is a part of."""
        return self._party
    
    @property
    def joined_at(self):
        """:class:`datetime.datetime`: The UTC time of when this member joined its party."""
        return self._joined_at

    @property
    def is_leader(self):
        """:class:`bool`: Returns ``True`` if member is the leader else ``False``."""
        return True if self.role else False

    @property
    def ready(self):
        """:class:`bool`: ``True`` if this member is ready else ``False``."""
        return self.meta.ready
    
    @property
    def input(self):
        """:class:`str`: The input type this user is currently using."""
        return self.meta.input

    @property
    def assisted_challenge(self):
        """:class:`str`: The current assisted challenge chosen by this member.
        ``None`` if no assisted challenge is set.
        """
        return self.meta.assisted_challenge

    @property
    def outfit(self):
        """:class:`str`: The CID of the outfit this user currently has equipped."""
        return self.meta.outfit

    @property
    def outfit_variants(self):
        """:class:`list`: A list containing the raw variants data for the currently equipped
        outfit.
        
        .. warning::
            
            Outfit variants doesn't seem to follow much logic. Therefore this returns the raw
            variants data received from fortnite's service. This can be directly passed with the
            ``variants`` keyword to :meth:`PartyMember.set_outfit()`.
        """
        return self.meta.outfit_variants

    @property
    def backpack(self):
        """:class:`str`: The BID of the backpack this member currently has equipped. 
        ``None`` if no backpack is equipped.
        """
        return self.meta.backpack
    
    @property
    def pickaxe(self):
        """:class:`str`: The pickaxe id of the pickaxe this member currently has equipped."""
        return self.meta.pickaxe

    @property
    def emote(self):
        """:class:`str`: The EID of the emote this member is currenyly playing.
        ``None`` if no emote is currently playing.
        """
        return self.meta.emote

    @property
    def banner(self):
        """:class:`tuple`: A tuple consisting of the icon id, color id and the season level.
        
        Example output: ::

            ('standardbanner15', 'defaultcolor15', 50)
        """
        return self.meta.banner

    @property
    def battlepass_info(self):
        """:class:`tuple`: A tuple consisting of has purchased, battlepass level, self boost xp, friends boost xp.
        
        Example output: ::
        
            (True, 30, 80, 70)    
        """
        return self.meta.battlepass_info
    
    @property
    def platform(self):
        """:class:`str`: The platform this user currently uses."""
        return self.meta.platform

    def _update(self, data):
        super()._update(data)
        self.role = data.get('role')
        self.revision = data.get('revision', 0)
        self.connections = data.get('connections', [])

    def update(self, data):
        if data['revision'] > self.revision:
            self.revision = data['revision']
        self.meta.update(data['member_state_updated'], raw=True)
        self.meta.remove(data['member_state_removed'])

    def update_role(self, role):
        self.role = role

    def create_variants(self, item="AthenaCharacter", particle_config='Emissive', **kwargs):
        """Creates the variants list by the variants you set.

        .. warning::

            This function is built upon data received from only some of the available outfits
            with variants. There is little logic behind the variants function therefore there
            might be some unexpected issues with this function. Please report such issues by
            creating an issue on the issue tracker or by reporting it to me on discord.
        
        Example usage: ::
        
            # set the outfit to soccer skin with Norwegian jersey and
            # the jersey number set to 99 (max number).
            async def set_soccer_skin():
                me = client.user.party.me

                variants = me.create_variants(
                    pattern=0,
                    numeric=99,
                    jersey_color='Norway'
                )

                await me.set_outfit(
                    asset='CID_149_Athena_Commando_F_SoccerGirlB',
                    variants=variants
                )
        
        Parameters
        ----------
        item: :class:`str`
            The variant item type. This defaults to ``AthenaCharacter`` which
            is what you want to use if you are changing skin variants.
        parts_config: :class:`str`
            The type of parts you want to use. The available types 
            are ``Emissive`` (default) and ``Mat``.
        pattern: Optional[:class:`int`]
            The pattern number you want to use.
        numeric: Optional[:class:`int`]
            The numeric number you want to use.
        clothing_color: Optional[:class:`int`]
            The clothing color you want to use.
        jersey_color: Optional[:class:`str`]
            The jersey color you want to use. For soccer skins this is the country
            you want the jersey to represent.
        parts: Optional[:class:`int`]
            The parts number you want to use.
        progressive: Optional[:class:`int`]
            The progressing number you want to use.
        particle: Optional[:class:`int`]
            The particle number you want to use.
        material: Optional[:class:`int`]
            The material number you want to use.
        emissive: Optional[:class:`int`]
            The emissive number you want to use.

        Returns
        -------
        List[:class:`dict`]
            List of dictionaries including all variants data.
        """
        config = {
            'pattern': 'Mat{}',
            'numeric': 'Numeric.{}',
            'clothing_color': 'Mat{}',
            'jersey_color': 'Color.{}',
            'parts': 'Stage{}',
            'progressive': 'Stage{}',
            'particle': '{}{}',
            'material': 'Mat{}',
            'emissive': 'Emissive{}'
        }

        variant = []
        for channel, value in kwargs.items():
            v = {
                'item': item,
                'channel': ''.join([x.capitalize() for x in channel.split('_')])
            }

            if channel == 'particle':
                v['variant'] = config[channel].format(particle_config, value)
            elif channel == 'JerseyColor':
                v['variant'] = config[channel].format(value.upper())
            else:
                v['variant'] = config[channel].format(value)
            variant.append(v)
        return variant


class PartyMember(PartyMemberBase):
    """Represents a party member.
    
    Attributes
    ----------
    client: :class:`Client`
        The client.
    """

    def __init__(self, client, party, data):
        super().__init__(client, party, data)

    async def kick(self):
        """|coro|
        
        Kicks this member from the party.

        Raises
        ------
        PartyPermissionError
            You are not the leader of the party.
        PartyError
            You attempted to kick yourself.
        HTTPException
            Something else went wrong when trying to kick this member.
        """
        if self.client.user.id != self.party.leader.id:
            raise PartyPermissionError(
                'You must be partyleader to perform this action')

        if self.client.user.id == self.id:
            raise PartyError('You can\'t kick yourself')

        await self.client.http.party_kick_member(self.party.id, self.id)
        self.party._remove_member(self.id)

    async def promote(self):
        """|coro|
        
        Promotes this user to partyleader.

        Raises
        ------
        PartyPermissionError
            You are not the leader of the party.
        PartyError
            You are already partyleader.
        HTTPException
            Something else went wrong when trying to promote this member.
        """
        if self.client.user.id != self.party.leader.id:
            raise PartyPermissionError(
                'You must be partyleader to perform this action')

        if self.client.user.id == self.id:
            raise PartyError('You are already partyleader')

        await self.client.http.party_promote_member(self.party.id, self.id)


class ClientPartyMember(PartyMemberBase):
    """Represents a the Clients party member object.
    
    Attributes
    ----------
    client: :class:`Client`
        The client.
    """

    def __init__(self, client, party, data):
        super().__init__(client, party, data)

        self.queue = asyncio.Queue(loop=self.client.loop)
        self.queue_active = False

    async def _patch(self, updated=None):
        meta = updated or self.meta.schema
        await self.client.http.party_update_member_meta(
            party_id=self.party.id,
            user_id=self.id,
            meta=meta,
            revision=self.revision
        )
        self.revision += 1
    
    async def patch(self, updated=None):
        future = self.client.loop.create_future()
        await self.queue.put((self._patch, future, {'updated': updated}))

        if not self.queue_active:
            asyncio.ensure_future(self._run_queue(), loop=self.client.loop)
        return await future

    async def _run_queue(self):
        self.queue_active = True
        try:
            while not self.queue.empty():
                func, future, kwargs = await self.queue.get()
                if func is None:
                    break

                while True:
                    try:
                        res = await func(**kwargs)
                        future.set_result(res)
                        break
                    except HTTPException as exc:
                        if exc.message_code == 'errors.com.epicgames.social.party.stale_revision':
                            self.revision = int(exc.message_vars[1])
                            continue
                        raise HTTPException(exc.response, exc.raw)
                    except FortniteException as exc:
                        future.set_exception(exc)
                        break

        except RuntimeError:
            pass
        self.queue_active = False

    async def leave(self):
        """|coro|
        
        Leaves the party.

        Raises
        ------
        HTTPException
            An error occured while requesting to leave the party.

        Returns
        -------
        :class:`Party`
            The new party the client is connected to after leaving.
        """
        await self.client.http.party_leave(self.party.id)
        self.client.xmpp.muc_room = None
        p = await self.client._create_party()
        self.client.user.set_party(p)
        return p

    async def set_ready(self, value):
        """|coro|
        
        Sets the readiness of the client.
        
        Parameters
        ----------
        value: :class:`bool`
            **True** to set it to ready.
            **False** to set it to unready.
        """
        prop = self.meta.set_readiness(
            val='Ready' if value is True else 'NotReady'
        )
        await self.patch(updated=prop)

    async def set_outfit(self, asset, key=None, variants=None):
        """|coro|
        
        Sets the outfit of the client.

        Parameters
        ----------
        asset: Required[:class:`str`]
            The CID of the outfit.

            .. note::

                You don't have to include the full path of the asset. The CID is
                enough.
        key: Optional[:class:`str`]
            The encyption key to use for this skin.
        variants: Optional[:class:`list`]
            The variants to use for this skin.
        """
        if '.' not in asset:
            asset = "AthenaCharacterItemDefinition'/Game/Athena/Items/Cosmetics/Characters/" \
                    "{0}.{0}'".format(asset)

        prop = self.meta.set_cosmetic_loadout(
            character=asset,
            character_ekey=key,
            variants=variants
        )
        await self.patch(updated=prop)
        
    async def set_backpack(self, asset, key=None):
        """|coro|
        
        Sets the backpack of the client.

        Parameters
        ----------
        asset: Required[:class:`str`]
            The CID of the backpack.

            .. note::

                You don't have to include the full path of the asset. The CID is
                enough.
        key: Optional[:class:`str`]
            The encyption key to use for this backpack.
        """
        if '.' not in asset:
            asset = "AthenaBackpackItemDefinition'/Game/Athena/Items/Cosmetics/Backpacks/" \
                    "{0}.{0}'".format(asset)

        prop = self.meta.set_cosmetic_loadout(
            backpack=asset,
            backpack_ekey=key
        )
        await self.patch(updated=prop)
    
    async def set_pickaxe(self, asset, key=None):
        """|coro|
        
        Sets the pickaxe of the client.

        Parameters
        ----------
        asset: Required[:class:`str`]
            The CID of the pickaxe.

            .. note::

                You don't have to include the full path of the asset. The CID is
                enough.
        key: Optional[:class:`str`]
            The encyption key to use for this pickaxe.
        """
        if '.' not in asset:
            asset = "AthenaPickaxeItemDefinition'/Game/Athena/Items/Cosmetics/Pickaxes/" \
                    "{0}.{0}'".format(asset)

        prop = self.meta.set_cosmetic_loadout(
            pickaxe=asset,
            pickaxe_ekey=key
        )
        await self.patch(updated=prop)

    async def set_emote(self, asset, run_for=None, key=None, section=None):
        """|coro|
        
        Sets the emote of the client.

        Parameters
        ----------
        asset: Required[:class:`str`]
            The EID of the emote.

            .. note::

                You don't have to include the full path of the asset. The EID is
                enough.
        run_for: Optional[:class:`int`]
            Seconds this emote should run for before being cancelled. ``None`` (default) means 
            will run infinitely and you can then clear it with :meth:`PartyMember.clear_emote()`.
        key: Optional[:class:`str`]
            The encyption key to use for this emote.
        section: Optional[:class:`int`]
            The section.
        """
        if '.' not in asset:
            asset = "AthenaDanceItemDefinition'/Game/Athena/Items/Cosmetics/Dances/" \
                    "{0}.{0}'".format(asset)

        prop = self.meta.set_emote(
            emote=asset,
            emote_ekey=key,
            section=section
        )
        await self.patch(updated=prop)

        if run_for is not None:
            self.client.loop.create_task(self._schedule_clear_emote(run_for))

    async def _schedule_clear_emote(self, seconds):
        await asyncio.sleep(seconds)
        await self.clear_emote()
    
    async def clear_emote(self):
        """|coro|
        
        Clears/stops the emote currently playing.
        """
        prop = self.meta.set_emote(
            emote='None',
            emote_ekey='',
            section=-1
        )
        await self.patch(updated=prop)
    
    async def set_banner(self, icon=None, color=None, season_level=None):
        """|coro|
        
        Sets the banner of the client.

        Parameters
        ----------
        icon: Optional[:class:`str`]
            The icon to use.
            *Defaults to standardbanner15*
        color: Optional[:class:`str`]
            The color to use.
            *Defaults to defaultcolor15*
        season_level: Optional[:class:`int`]
            The season level.
            *Defaults to 1*
        """
        prop = self.meta.set_banner(
            banner_icon=icon,
            banner_color=color,
            season_level=season_level
        )
        await self.patch(updated=prop)
    
    async def set_battlepass_info(self, has_purchased=None, level=None, self_boost_xp=None,
                                  friend_boost_xp=None):
        """|coro|
        
        Sets the battlepass info of the client.

        .. note::

            This is simply just for showing off. It just shows visually so
            boostxp, level and stuff will not work, just show.

        Parameters
        ----------
        has_purchased: Optional[:class:`bool`]
            Shows visually that you have purchased the battlepass.
            *Defaults to False*
        level: Optional[:class:`int`]
            Sets the level and shows it visually.
            *Defaults to 1*
        self_boost_xp: Optional[:class:`int`]
            Sets the self boost xp and shows it visually.
        friend_boost_xp: Optional[:class:`int`]
            Set the friend boost xp and shows it visually.
        """
        prop = self.meta.set_battlepass_info(
            has_purchased=has_purchased,
            level=level,
            self_boost_xp=self_boost_xp,
            friend_boost_xp=friend_boost_xp
        )
        await self.patch(updated=prop)
    
    async def set_assisted_challenge(self, quest=None, num_completed=None):
        """|coro|
        
        Sets the assisted challenge.
        
        Parameters
        ----------
        quest: :class:`str`
            The quest to set.

            .. note::

                You don't have to include the full path of the quest. The quest id is
                enough.
        num_completed: :class:`int`
            How many quests you have completed, I think (didn't test this).
        """
        if '.' not in quest:
            quest = "FortQuestItemDefinition'/Game/Athena/Items/Quests/DailyQuests/Quests/" \
                    "{0}.{0}'".format(quest)

        prop = self.meta.set_assisted_challenge(
            quest=quest,
            completed=num_completed
        )
        await self.patch(updated=prop)    


class PartyBase:

    def __init__(self, client, data):
        self._client = client
        self._id = data.get('id')
        self._members = {}
        self._applicants = data.get('applicants', [])

        self._update_invites(data.get('invites', []))
        self._update_config(data.get('config'))
        self.meta = PartyMeta(self, data['meta'])

    @property
    def client(self):
        """:class:`Client`: The client."""
        return self._client

    @property
    def id(self):
        """:class:`str`: The party's id."""
        return self._id

    @property
    def members(self):
        """:class:`dict`: Mapping of the party's members. *Example: {memberid: :class:`PartyMember`}*"""
        return self._members

    @property
    def member_count(self):
        """:class:`int`: The amount of member currently in this party."""
        return len(self._members)

    @property
    def applicants(self):
        """:class:`list`: The party's applicants."""
        return self._applicants

    @property
    def leader(self):
        """:class:`PartyMember`: The leader of the party."""
        for member in self.members.values():
            if member.is_leader:
                return member

    @property
    def playlist_info(self):
        """:class:`tuple`: A tuple containing the name, tournament, event window and region
        of the currently set playlist.
        
        Example output: ::

            # output for default duos
            ('Playlist_DefaultDuo', '', '', 'EU')

            # output for arena trios
            ('Playlist_ShowdownAlt_Trios', 'epicgames_Arena_S10_Trios', 'Arena_S10_Division1_Trios', 'EU)
        """
        return self.meta.playlist_info

    @property
    def squad_fill_enabled(self):
        """:class:`bool`: ``True`` if squad fill is enabled else ``False``."""
        return self.meta.squad_fill_enabled

    @property
    def privacy(self):
        """:class:`PartyPrivacy`: The currently set privacy of this party."""
        return self.meta.privacy

    def _add_member(self, member):
        self.members[member.id] = member

    def _remove_member(self, id):
        if not isinstance(id, str):
            id = id.id
        del self.members[id]

    def _update(self, data):
        try:
            config = data['config']
        except KeyError:
            config = {
                'joinability': data['party_privacy_type'],
                'max_size': data['max_number_of_members'],
                'sub_type': data['party_sub_type'],
                'type': data['party_type'],
                'invite_ttl_seconds': data['invite_ttl_seconds']
            }

        self._update_config({**self.config, **config})

        self.meta.update(data['party_state_updated'], raw=True)
        self.meta.remove(data['party_state_removed'])

        privacy = self.meta.get_prop('PrivacySettings_j')
        _p = privacy['PrivacySettings']
        found = False
        for d in PartyPrivacy:
            p = d.value
            if p['partyType'] != _p['partyType']:
                continue
            if p['inviteRestriction'] != _p['partyInviteRestriction']:
                continue
            if p['onlyLeaderFriendsCanJoin'] != _p['bOnlyLeaderFriendsCanJoin']:
                continue
            found = p
            break

        if found:
            self.config['privacy'] = found

    def _update_invites(self, invites):
        self.invites = invites

    def _update_config(self, config=None):
        config = config if config is not None else {}

        self.join_confirmation = config['join_confirmation']
        self.max_size = config['max_size']
        self.invite_ttl_seconds = config.get('invite_ttl_seconds', config['invite_ttl'])
        self.sub_type = config['sub_type']
        self.config = {**self.client.default_party_config, **config}

    async def _update_members(self, raw_members):
        for raw in raw_members:
            user_id = raw.get('account_id', raw.get('accountId'))
            if user_id == self.client.user.id:
                user = self.client.user
            else:
                user = self.client.get_user(user_id)
                if user is None:
                    user = await self.client.fetch_profile(user_id)
            raw = {**raw, **(user.get_raw())}

            member = PartyMember(self.client, self, raw)
            self._add_member(member)

        ids = map(lambda r: raw.get('account_id', raw.get('accountId')), raw_members)
        to_remove = []
        for m in self.members.values():
            if m.id not in ids:
                to_remove.append(m.id)

        for id in to_remove:
            self._remove_member(id)

    async def _update_members_meta(self):
        data = await self.client.http.party_lookup(self.id)
        for m in data['members']:
            member = self.members[m['account_id']]
            member.meta.update(m['meta'], raw=True)


class Party(PartyBase):
    """Represent a party that the ClientUser is not yet a part of."""

    def __init__(self, client, data):
        super().__init__(client, data)


class ClientParty(PartyBase):
    """Represents ClientUser's party."""

    def __init__(self, client, data):
        super().__init__(client, data)

        self.last_raw_status = None
        self._me = None

        self.queue = asyncio.Queue(loop=self.client.loop)
        self.queue_active = False
        
        self._update_revision(data.get('revision', 0))
        self._update_invites(data.get('invites', []))
        self._update_config(data.get('config'))
        self.meta = PartyMeta(self, data['meta'])

    @property
    def me(self):
        """:class:`ClientPartyMember`: The clients partymember object."""
        return self._me

    @property
    def muc_jid(self):
        """:class:`aioxmpp.JID`: The JID of the party MUC."""
        return aioxmpp.JID.fromstr('Party-{}@muc.prod.ol.epicgames.com'.format(self.id))

    def _add_clientmember(self, member):
        self._me = member

    def _remove_member(self, id):
        if not isinstance(id, str):
            id = id.id
        del self.members[id]
        self.update_presence()

    def update_presence(self, text=None, conf={}):
        perm = self.config['privacy']['presencePermission']
        if perm == 'Noone' or (perm == 'Leader' and (self.me is not None and not self.me.is_leader)):
            join_data = {
                'bInPrivate': True
            }
        else:
            join_data = {
                'sourceId': self.client.user.id,
                'sourceDisplayName': self.client.user.display_name,
                'sourcePlatform': self.client.platform,
                'partyId': self.id,
                'partyTypeId': 286331153,
                'key': 'k',
                'appId': 'Fortnite',
                'buildId': self.client.net_cl,
                'partyFlags': -2024557306,
                'notAcceptingReason': 0,
                'pc': len(self.members.keys()),
            }

        _default_status = {
            'Status': 'Battle Royale Lobby - {0} / {1}'.format(len(self.members.keys()), self.max_size),
            'bIsPlaying': True,
            'bIsJoinable': False,
            'bHasVoiceSupport': False,
            'SessionId': '',
            'Properties': {
                'party.joininfodata.286331153_j': join_data,
                'FortBasicInfo_j': {
                    'homeBaseRating': 1,
                },
                'FortLFG_I': '0',
                'FortPartySize_i': 1,
                'FortSubGame_i': 1,
                'InUnjoinableMatch_b': False,
                'FortGameplayStats_j': {
                    'state': '',
                    'playlist': 'None',
                    'numKills': 0,
                    'bFellToDeath': False,
                },
                'GamePlaylistName_s': self.meta.playlist_info[0],
                'Event_PlayersAlive_s': '0',
                'Event_PartySize_s': str(len(self.members)),
                'Event_PartyMaxSize_s': str(self.max_size),
            },
        }

        if text is None:
            if self.client.status is None:
                _text = {}
            else:
                _text = {'Status': str(self.client.status)}
        else:
            _text = {'Status': str(text)}
        
        if self.client.status is not False:
            self.last_raw_status = {**_default_status, **conf, **_text}
            self.client.xmpp.set_presence(status=self.last_raw_status)
        
    def _update(self, data):
        if self.revision < data['revision']:
            self.revision = data['revision']

        try:
            config = data['config']
        except KeyError:
            config = {
                'joinability': data['party_privacy_type'],
                'max_size': data['max_number_of_members'],
                'sub_type': data['party_sub_type'],
                'type': data['party_type'],
                'invite_ttl_seconds': data['invite_ttl_seconds']
            }
        
        self._update_config({**self.config, **config})

        self.meta.update(data['party_state_updated'], raw=True)
        self.meta.remove(data['party_state_removed'])

        privacy = self.meta.get_prop('PrivacySettings_j')
        _p = privacy['PrivacySettings']
        found = False
        for d in PartyPrivacy:
            p = d.value
            if p['partyType'] != _p['partyType']:
                continue
            if p['inviteRestriction'] != _p['partyInviteRestriction']:
                continue
            if p['onlyLeaderFriendsCanJoin'] != _p['bOnlyLeaderFriendsCanJoin']:
                continue
            found = p
            break
        
        if found:
            self.config['privacy'] = found
        
        if self.client.status is not False:
            self.update_presence()

    def _update_revision(self, revision):
        self.revision = revision

    def _update_config(self, config=None):
        config = config if config is not None else {} 
        
        self.join_confirmation = config['join_confirmation']
        self.max_size = config['max_size']
        self.invite_ttl_seconds = config.get('invite_ttl_seconds', config['invite_ttl'])
        self.sub_type = config['sub_type']
        self.config = {**self.client.default_party_config, **config}

    async def _update_members(self, raw_members):
        for raw in raw_members:
            user_id = raw.get('account_id', raw.get('accountId'))
            if user_id == self.client.user.id:
                user = self.client.user
            else:
                user = self.client.get_user(user_id)
                if user is None:
                    user = await self.client.fetch_profile(user_id)
            raw = {**raw, **(user.get_raw())}

            member = PartyMember(self.client, self, raw)
            self._add_member(member)
        
            if member.id == self.client.user.id:
                clientmember = ClientPartyMember(self.client, self, raw)
                self._add_clientmember(clientmember)

        ids = map(lambda r: raw.get('account_id', raw.get('accountId')), raw_members)
        to_remove = []
        for m in self.members.values():
            if m.id not in ids:
                to_remove.append(m.id)
            
        for id in to_remove:
            self._remove_member(id)

    async def join_chat(self):
        await self.client.xmpp.join_muc(self.id)
    
    async def send(self, content):
        """|coro|
        
        Sends a message to this party's chat.

        Parameters
        ----------
        content: :class:`str`
            The content of the message.
        """
        await self.client.xmpp.send_party_message(content)

    async def _patch(self, updated=None, deleted=None):
        await self.client.http.party_update_meta(
            party_id=self.id,
            updated_meta=updated or self.meta.schema,
            deleted_meta=deleted or [],
            config=self.config,
            revision=self.revision
        )
        self.revision += 1
    
    async def patch(self, updated=None, deleted=None):
        future = self.client.loop.create_future()
        await self.queue.put((self._patch, future, {'updated': updated, 'deleted': deleted}))

        if not self.queue_active:
            asyncio.ensure_future(self._run_queue(), loop=self.client.loop)
        return await future

    async def _run_queue(self):
        self.queue_active = True
        try:
            while not self.queue.empty():
                func, future, kwargs = await self.queue.get()
                if func is None:
                    break
                
                while True:
                    try:
                        res = await func(**kwargs)
                        future.set_result(res)
                        break
                    except HTTPException as exc:
                        if exc.message_code == 'errors.com.epicgames.social.party.stale_revision':
                            self.revision = int(exc.message_vars[1])
                            continue
                        raise HTTPException(exc.response, exc.raw)
                    except FortniteException as exc:
                        future.set_exception(exc)
                        break

        except RuntimeError:
            pass
        self.queue_active = False
    
    async def invite(self, user_id):
        """|coro|
        
        Invites a user to the party.

        Parameters
        ----------
        user_id: :class:`str`
            The id of the user to invite.

        Raises
        ------
        PartyError
            The party is full.
        HTTPException
            Something else went wrong when trying to invite the user.
        """
        if len(self.members.keys()) == self.max_size:
            raise PartyError('Party is full')
        
        await self.client.http.party_send_invite(user_id)

    async def _leave(self):
        """|coro|
        
        Leaves the party.
        
        Raises
        ------
        HTTPException
            Something went wrong when trying to leave the party.
        """
        self.client.xmpp.muc_room = None
        await self.client.http.party_leave(self.id)

    async def set_privacy(self, privacy):
        """|coro|
        
        Sets the privacy of the party.
        
        Parameters
        ----------
        privacy: :class:`.PartyPrivacy`

        Raises
        ------
        PartyPermissionError
            The client is not the leader of the party.
        """
        if self.leader.id != self.client.user.id:
            raise PartyPermissionError('You have to be leader for this action to work.')

        if not isinstance(privacy, dict):
            privacy = privacy.value

        updated, deleted = self.meta.set_privacy(privacy)
        await self.patch(updated=updated, deleted=deleted)
    
    async def set_playlist(self, playlist=None, tournament=None, event_window=None, region=None):
        """|coro|
        
        Sets the current playlist of the party.

        Sets the playlist to Duos EU: ::

            await party.set_playlist(
                playlist='Playlist_DefaultDuo',
                region='EU'
            )
        
        Sets the playlist to Arena Trios EU (Replace ``Trios`` with ``Solo`` for arena solo): ::

            await party.set_playlist(
                playlist='Playlist_ShowdownAlt_Trios',
                tournament='epicgames_Arena_S10_Trios',
                event_window='Arena_S10_Division1_Trios',
                region='EU'
            )

        Parameters
        ----------
        playlist: Optional[:class:`str`]
            The name of the playlist.
            *Defaults to 'EU'*
        tournament: Optional[:class:`str`]
            The tournament id.
        event_window: Optional[:class:`str`]
            The event window id.
        region: Optional[:class:`str`]
            The region to use.
            *Defaults to 'EU'*

        Raises
        ------
        PartyPermissionError
            The client is not the leader of the party.
        """
        if self.leader.id != self.client.user.id:
            raise PartyPermissionError('You have to be leader for this action to work.')

        prop = self.meta.set_playlist(
            playlist=playlist,
            tournament=tournament,
            event_window=event_window,
            region=region
        )
        await self.patch(updated=prop)
    
    async def set_custom_key(self, key):
        """|coro|
        
        Sets the custom key of the party.

        Parameters
        ----------
        key: :class:`str`
            The key to set.

        Raises
        ------
        PartyPermissionError
            The client is not the leader of the party.
        """
        if self.leader.id != self.client.user.id:
            raise PartyPermissionError('You have to be leader for this action to work.')

        prop = self.meta.set_custom_key(
            key=key
        )
        await self.patch(updated=prop)

    async def set_fill(self, value):
        """|coro|
        
        Sets the fill status of the party.

        Parameters
        ----------
        value: :class:`bool`
            What to set the fill status to.

            **True** sets it to 'Fill'
            **False** sets it to 'NoFill'

        Raises
        ------
        PartyPermissionError
            The client is not the leader of the party.
        """
        if self.leader.id != self.client.user.id:
            raise PartyPermissionError('You have to be leader for this action to work.')

        prop = self.meta.set_fill(val=value)
        await self.patch(updated=prop)


class PartyInvitation:
    """Represents a party invitation.
    
    Attributes
    ----------
    client: :class:`Client`
        The client.
    party: :class:`Party`
        The party the invitation belongs to.
    author: :class:`Friend`
        The friend that invited you to the party.
    created_at: :class:`datetime.datetime`
        The UTC time this invite was created at.
    """
    def __init__(self, client, party, data):
        self.client = client
        self.party = party
        
        self.author = self.client.get_friend(data['sent_by'])
        self.created_at = self.client.from_iso(data['sent_at'])

    async def accept(self):
        """|coro|
        
        Accepts the invitation and joins the party.

        Raises
        ------
        HTTPException
            Something went wrong when accepting the invitation.
        """
        await self.client.join_to_party(self.party.id)

    async def decline(self):
        """|coro|
        
        Declines the invitation.

        Raises
        ------
        HTTPException
            Something went wrong when declining the invitation.
        """
        await self.client.http.party_decline_invite(self.party.id)


class PartyJoinConfirmation:
    """Represents a join confirmation.
    
    Attributes
    ----------
    client: :class:`Client`
        The client.
    party: :class:`Party`
        The party the user wants to join.
    user: :class:`User`
        The user who requested to join the party.
    created_at: :class:`datetime.datetime`
        The UTC time of when the join confirmation was received.
    """
    def __init__(self, client, party, data):
        self.client = client
        self.party = party
        self.user = User(self.client, data)
        self.created_at = self.client.from_iso(data['sent'])

    async def confirm(self):
        """|coro|
        
        Confirms this user.

        Raises
        ------
        HTTPException
            Something went wrong when confirming this user.
        """
        await self.client.http.party_member_confirm(self.party.id, self.user.id)

    async def reject(self):
        """|coro|
        
        Rejects this user.

        Raises
        ------
        HTTPException
            Something went wrong when rejecting this user.
        """
        await self.client.http.party_member_reject(self.party.id, self.user.id)
 



# -*- coding: utf-8 -*-

"""
MIT License

Copyright (c) 2019 Terbau

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import aiohttp
import asyncio
import logging
import json

from .errors import HTTPException

log = logging.getLogger(__name__)


class HTTPClient:
    def __init__(self, client, connector=None):
        self.client = client
        self.connector = connector
        self._jar = aiohttp.CookieJar()
        self.headers = {}

        self.create_connection()
    
    @classmethod
    async def json_or_text(cls, response):
        text = await response.text(encoding='utf-8')
        if response.headers.get('content-type') == 'application/json':
            return json.loads(text)
        return text

    @property
    def launcher_user_agent(self):
        return f''

    @property
    def fortnite_user_agent(self):
        return f''

    def get_auth(self, auth):
        s = 'basic '
        if auth.upper() == 'FORTNITE':
            s += self.client.fortnite_token
        elif auth.upper() == 'LAUNCHER':
            s += self.client.launcher_token
        else:
            s = auth
        
        return s

    def add_header(self, key, val):
        self.headers[key] = val

    def remove_header(self, key):
        return self.headers.pop(key)

    async def close(self):
        self._jar.clear()
        if self.__session:
            await self.__session.close()

    def create_connection(self):
        self.__session = aiohttp.ClientSession(
            connector=self.connector,
            loop=self.client.loop,
            cookie_jar=self._jar
        )

    async def request(self, method, url, is_json=False, **kwargs):
        headers = {**kwargs.get('headers', {}), **self.headers}

        if is_json:
            try:
                kwargs['data'] = json.dumps(kwargs['data'])
            except KeyError:
                pass
            headers['content-type'] = 'application/json'

        if headers != {}:
            kwargs['headers'] = headers

        async with self.__session.request(method, url, **kwargs) as r:
            log.debug(f'{method} {url} has returned {r.status}')
            data = await self.json_or_text(r)
            try:
                _data = json.loads(data)
                if 'errorCode' in _data.keys():
                    raise HTTPException(r, _data)
            except (KeyError, TypeError, json.decoder.JSONDecodeError):
                pass
            return data

    async def fn_request(self, method, url, auth, **kwargs):
        headers = kwargs.get('headers', {})
        headers['User-Agent'] = 'EpicGamesLauncher/9.6.1-4858958+++Portal+Release-Live Windows/10.0.17134.1.768.64bit'
        headers['Authorization'] = self.get_auth(auth)
        kwargs['headers'] = headers
        
        return await self.request(method, url, **kwargs)

    async def get(self, url, auth, **kwargs):
        return await self.fn_request('GET', url, auth, **kwargs)

    async def post(self, url, auth, **kwargs):
        return await self.fn_request('POST', url, auth, **kwargs)

    async def delete(self, url, auth, **kwargs):
        return await self.fn_request('DELETE', url, auth, **kwargs)

    async def patch(self, url, auth, **kwargs):
        return await self.fn_request('PATCH', url, auth, **kwargs)

    async def get_profile_by_display_name(self, display_name):
        return await self.get(
            'https://account-public-service-prod03.ol.epicgames.com/account/api/' \
            'public/account/displayName/{0}'.format(display_name),
            self.client.auth.authorization
        )

    async def get_profile(self, user_id):
        return await self.get(
            'https://account-public-service-prod03.ol.epicgames.com/account/api/' \
            'public/account/{0}'.format(user_id),
            self.client.auth.authorization
        )

    async def get_profiles(self, user_ids):
        accounts = ['&accountId={}'.format(user_id) for user_id in user_ids]
        return await self.get(
            'https://account-public-service-prod03.ol.epicgames.com/account/api/' \
            'public/account?{0}'.format(''.join(accounts)),
            self.client.auth.authorization
        )
    
    async def get_friends(self, include_pending=False):
        _incl = True if include_pending else False
        return await self.get(
            'https://friends-public-service-prod06.ol.epicgames.com/friends/api/' \
            'public/friends/{0.client.user.id}?includePending={1}'.format(self, _incl),
            self.client.auth.authorization
        )

    async def add_friend(self, user_id):
        return await self.post(
            'https://friends-public-service-prod06.ol.epicgames.com/friends/api/' \
            'public/friends/{0.client.user.id}/{1}'.format(self, user_id),
            self.client.auth.authorization
        )

    async def remove_friend(self, user_id):
        return await self.delete(
            'https://friends-public-service-prod06.ol.epicgames.com/friends/api/' \
            'public/friends/{0.client.user.id}/{1}'.format(self, user_id),
            self.client.auth.authorization
        )

    async def get_friends_blocklist(self):
        return await self.get(
            'https://friends-public-service-prod06.ol.epicgames.com/friends/api/' \
            'public/blocklist/{0.client.user.id}'.format(self),
            self.client.auth.authorization
        )

    async def block_user(self, user_id):
        return await self.post(
            'https://friends-public-service-prod06.ol.epicgames.com/friends/api/' \
            'public/blocklist/{0.client.user.id}/{1}'.format(self, user_id),
            self.client.auth.authorization
        )

    async def unblock_user(self, user_id):
        return await self.delete(
            'https://friends-public-service-prod06.ol.epicgames.com/friends/api/' \
            'public/blocklist/{0.client.user.id}/{1}'.format(self, user_id),
            self.client.auth.authorization
        )

    async def get_br_stats_v1(self, user_id):
        return await self.get(
            'https://fortnite-public-service-prod11.ol.epicgames.com/fortnite/api/' \
            'stats/{0}'.format(user_id),
            self.client.auth.authorization
        )

    async def get_br_stats_v2(self, user_id, start_time=None, end_time=None):
        query_parameters = ''
        if start_time:
            query_parameters += '&startTime={}'.format(start_time)
        if end_time:
            query_parameters += '&endTime={}'.format(end_time)

        return await self.get(
            'https://fortnite-public-service-prod11.ol.epicgames.com/fortnite/api/' \
            'statsv2/account/{0}{1}'.format(user_id, query_parameters),
            self.client.auth.authorization
        )

    async def get_multiple_br_stats_v2(self, ids, stats, start_time=None, end_time=None):
        params = {}
        if start_time:
            params['startTime'] = start_time
        if end_time:
            params['endTime'] = end_time

        return await self.post(
            'https://fortnite-public-service-prod11.ol.epicgames.com/fortnite/api/statsv2/query',
            self.client.auth.authorization,
            data={
                'owners': ids,
                'stats': stats
            },
            is_json=True,
            params=params
        )

    async def get_br_leaderboard(self, stat, platform, mode, window, page=0, items_per_page=50):
        return await self.post(
            'https://fortnite-public-service-prod11.ol.epicgames.com/fortnite/api/' \
            'leaderboards/type/global/stat/br_{0}_{1}_m0_{2}/window/{3}?ownertype=1'.format(
                stat,
                platform.value,
                mode.value,
                window.value
            ),
            self.client.auth.authorization,
            is_json=True,
            params={
                'pageNumber': page,
                'itemsPerPage': items_per_page
            }
        )

    async def get_br_leaderboard_v2(self, stat):
        return await self.get(
            'https://fortnite-public-service-prod11.ol.epicgames.com/fortnite/api/' \
            'statsv2/leaderboards/{0}'.format(stat),
            self.client.auth.authorization
        )

    async def get_lightswitch_status(self, service_id=None):
        return await self.get(
            'https://lightswitch-public-service-prod06.ol.epicgames.com/lightswitch/api/' \
            'service/bulk/status',
            self.client.auth.authorization,
            params={'serviceId': service_id} if service_id else None
        )

    async def get_store_catalog(self):
        return await self.get(
            'https://fortnite-public-service-prod11.ol.epicgames.com/fortnite/api/storefront/v2/catalog',
            self.client.auth.authorization
        )

    async def get_fortnite_news(self):
        return await self.get(
            'https://fortnitecontent-website-prod07.ol.epicgames.com/content/api/pages/fortnite-game',
            self.client.auth.authorization
        )

    async def get_fortnite_content(self):
        data = await self.get(
            'https://fortnitecontent-website-prod07.ol.epicgames.com/content/api/pages/fortnite-game',
            self.client.auth.authorization
        )
        return json.loads(data)

    async def get_fortnite_timeline(self):
        return await self.get(
            'https://fortnite-public-service-prod11.ol.epicgames.com/fortnite/api/calendar/v1/timeline',
            self.client.auth.authorization
        )

    ########################################
    # Party
    ########################################

    # send party invite
    async def party_send_invite(self, user_id):
        return await self.post(
            'https://party-service-prod.ol.epicgames.com/party/api/v1/Fortnite/' \
            'user/{0}/pings/{1.client.user.id}'.format(user_id, self),
            self.client.auth.authorization,
            is_json=True
        )

    # partyinvite decline
    async def party_decline_invite(self, party_id):
        return await self.post(
            'https://party-service-prod.ol.epicgames.com/party/api/v1/Fortnite/' \
            'parties/{0}/invites/{1.client.user.id}/decline'.format(party_id, self),
            self.client.auth.authorization,
            is_json=True
        )

    async def party_member_confirm(self, party_id, user_id):
        return await self.post(
            'https://party-service-prod.ol.epicgames.com/party/api/v1/Fortnite/' \
            'parties/{0}/members/{1}/confirm'.format(party_id, user_id),
            self.client.auth.authorization,
            is_json=True

        )
    
    async def party_member_reject(self, party_id, user_id):
        return await self.post(
            'https://party-service-prod.ol.epicgames.com/party/api/v1/Fortnite/' \
            'parties/{0}/members/{1}/reject'.format(party_id, user_id),
            self.client.auth.authorization,
            is_json=True
        )

    async def party_promote_member(self, party_id, user_id):
        return await self.post(
            'https://party-service-prod.ol.epicgames.com/party/api/v1/Fortnite/' \
            'parties/{0}/members/{1}/promote'.format(party_id, user_id),
            self.client.auth.authorization,
            is_json=True
        )

    async def party_kick_member(self, party_id, user_id):
        return await self.delete(
            'https://party-service-prod.ol.epicgames.com/party/api/v1/Fortnite/' \
            'parties/{0}/members/{1}'.format(party_id, user_id),
            self.client.auth.authorization,
        )

    async def party_leave(self, party_id):
        data = {
            'connection': {
                'id': self.client.user.jid,
                'meta': {
                    'urn:epic:conn:platform_s': self.client.platform,
                    'urn:epic:conn:type_s': 'game'
                }
            },
            'meta': {
                'urn:epic:member:dn_s': self.client.user.display_name,
                'urn:epic:member:type_s': 'game',
                'urn:epic:member:platform_s': self.client.platform,
                'urn:epic:member:joinrequest_j': '{"CrossplayPreference_i":"1"}',
            }
        }

        return await self.delete(
            'https://party-service-prod.ol.epicgames.com/party/api/v1/Fortnite/' \
            'parties/{0}/members/{1.client.user.id}'.format(party_id, self),
            self.client.auth.authorization,
            data=data,
            is_json=True
        )

    async def party_join_request(self, party_id):
        data = {
            'connection': {
                'id': str(self.client.xmpp.xmpp_client.local_jid),
                'meta': {
                    'urn:epic:conn:platform_s': self.client.platform,
                    'urn:epic:conn:type_s': 'game',
                },
            },
            'meta': {
                'urn:epic:member:dn_s': self.client.user.display_name,
                'urn:epic:member:type_s': 'game',
                'urn:epic:member:platform_s': self.client.platform,
                'urn:epic:member:joinrequest_j': '{"CrossplayPreference_i":"1","SubGame_u":"1"}',
            },
        }

        return await self.post(
            'https://party-service-prod.ol.epicgames.com/party/api/v1/Fortnite/' \
            'parties/{0}/members/{1.client.user.id}/join'.format(party_id, self),
            self.client.auth.authorization,
            data=data,
            is_json=True
        )
    
    async def party_lookup(self, party_id):
        return await self.get(
            'https://party-service-prod.ol.epicgames.com/party/api/v1/Fortnite/' \
            'parties/{0}'.format(party_id),
            self.client.auth.authorization
        )
    
    async def party_lookup_user(self, user_id):
        return await self.get(
            'https://party-service-prod.ol.epicgames.com/party/api/v1/Fortnite/' \
            'user/{0}'.format(user_id),
            self.client.auth.authorization
        )

    async def party_create(self, config):
        data = {
            'config': {
                'join_confirmation': config['join_confirmation'],
                'joinability': config['joinability'],
                'max_size': config['max_size']
            },
            'join_info': {
                'connection': {
                    'id': str(self.client.xmpp.xmpp_client.local_jid),
                    'meta': {
                        'urn:epic:conn:platform_s': self.client.platform,
                        'urn:epic:conn:type_s': 'game'
                    }
                }
            },
            'meta': {
                'urn:epic:cfg:party-type-id_s': 'default',
                'urn:epic:cfg:build-id_s': str(self.client.party_build_id),
                'urn:epic:cfg:join-request-action_s': 'Manual',
                'urn:epic:cfg:chat-enabled_b': str(config['chat_enabled']).lower()
            }
        }

        return await self.post(
            'https://party-service-prod.ol.epicgames.com/party/api/v1/Fortnite/' \
            'parties',
            self.client.auth.authorization,
            data=data,
            is_json=True
        )

    async def party_update_member_meta(self, party_id, user_id, meta, revision):
        data = {
            'delete': [],
            'revision': revision,
            'update': meta
        }

        await self.patch(
            'https://party-service-prod.ol.epicgames.com/party/api/v1/Fortnite/' \
            'parties/{0}/members/{1}/meta'.format(party_id, user_id),
            self.client.auth.authorization,
            data=data,
            is_json=True
        )

    async def party_update_meta(self, party_id, updated_meta, deleted_meta, config, revision):
        data = {
            'config': {
                'join_confirmation': config['join_confirmation'],
                'joinability': config['joinability'],
                'max_size': config['max_size']
            },
            'meta': {
                'delete': deleted_meta,
                'update': updated_meta
            },
            'party_state_overridden': {},
            'party_privacy_type': config['joinability'],
            'party_type': config['type'],
            'party_sub_type': config['sub_type'],
            'max_number_of_members': config['max_size'],
            'invite_ttl_seconds': config['invite_ttl_seconds'],
            'revision': revision
        }

        return await self.patch(
            'https://party-service-prod.ol.epicgames.com/party/api/v1/Fortnite/' \
            'parties/{0}'.format(party_id),
            self.client.auth.authorization,
            data=data,
            is_json=True
        )



# -*- coding: utf-8 -*-

"""
MIT License

Copyright (c) 2019 Terbau

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import datetime
import asyncio
import logging
import traceback
import uuid

from .errors import AuthException, HTTPException

log = logging.getLogger(__name__)


class Auth:
    def __init__(self, client):
        self.client = client
        self.device_id = self.client.device_id or uuid.uuid4().hex

    @property
    def launcher_authorization(self):
        return 'bearer {0}'.format(self.launcher_access_token)

    @property
    def authorization(self):
        return 'bearer {0}'.format(self.access_token)

    async def authenticate(self):
        try:
            data = {
                'grant_type': 'password',
                'username': self.client.email,
                'password': self.client.password
            }

            try:
                data = await self.grant_session('LAUNCHER', data=data)
            except HTTPException as exc:
                log.debug('Could not authenticate normally, checking why now.')

                if exc.message_code != 'errors.com.epicgames.common.two_factor_authentication.required':
                    raise HTTPException(exc.response, exc.raw)
                
                log.debug('2fa code required to continue login process. Asking for code now.')

                code = self.client.two_factor_code
                if code is None:
                    code = int(input('Please enter the 2fa code:\n'))

                data = {
                    'grant_type': 'otp',
                    'otp': str(code),
                    'challenge': exc.raw['challenge']
                }
                data = await self.grant_session('LAUNCHER', data=data)
                log.debug('Valid 2fa code entered')

            self.launcher_access_token = data['access_token']
            await self.exchange_code('bearer {0}'.format(self.launcher_access_token))
        except asyncio.CancelledError:
            pass
        except Exception as e:
            traceback.print_exc()
            raise AuthException('Could not authenticate. Error: {}'.format(e))

    async def grant_refresh_token(self, refresh_token):
        data = {
            'grant_type': 'refresh_token',
            'refresh_token': refresh_token
        }
        return await self.grant_session('FORTNITE', data=data)

    async def get_exchange_code(self, auth):
        res = await self.client.http.get(
            'https://account-public-service-prod03.ol.epicgames.com/account/api/oauth/exchange',
            auth
        )
        return res.get('code')

    async def exchange_code(self, auth):
        code = await self.get_exchange_code(auth)
        if code is None:
            raise AuthException('Could not get exchange code')

        data = {
            'grant_type': 'exchange_code',
            'token_type': 'eg1',
            'exchange_code': code
        }
        
        data = await self.grant_session('FORTNITE', data=data)
        self._update(data)

    async def grant_session(self, auth, **kwargs):
        headers = {
            'X-Epic-Device-ID': self.device_id
        }
        data = await self.client.http.post(
            'https://account-public-service-prod03.ol.epicgames.com/account/api/oauth/token',
            auth,
            **{**kwargs, 'headers': headers}
        )
        return data

    async def kill_token(self, token):
        await self.client.http.delete(
            'https://account-public-service-prod03.ol.epicgames.com/account/api/' \
            'oauth/sessions/kill/{0}'.format(token),
            self.authorization
        )

    async def kill_other_sessions(self):
        await self.client.http.delete(
            'https://account-public-service-prod03.ol.epicgames.com/account/api/' \
            'oauth/sessions/kill?killType=OTHERS_ACCOUNT_CLIENT_SERVICE',
            self.authorization
        )

    async def get_eula_version(self, account_id):
        res = await self.client.http.get(
            'https://eulatracking-public-service-prod-m.ol.epicgames.com/eulatracking/' \
            'api/public/agreements/fn/account/{0}'.format(account_id),
            self.authorization # MIGHT BE fn auth
        )
        return res['version'] if isinstance(res, dict) else 0

    async def _accept_eula(self, version, account_id):
        await self.client.http.post(
            'https://eulatracking-public-service-prod-m.ol.epicgames.com/eulatracking/' \
            'api/public/agreements/fn/version/{0}/account/{1}/accept?locale=en'.format(
                version, 
                account_id
            ),
            self.authorization # MIGHT BE fn auth
        )

    async def accept_eula(self, account_id):
        version = await self.get_eula_version(account_id)
        if version != 0:
            await self._accept_eula(version, account_id)
            await self._grant_access(account_id)

    async def _grant_access(self, account_id):
        await self.client.http.post(
            'https://fortnite-public-service-prod11.ol.epicgames.com/fortnite/api/' \
            'game/v2/grant_access/{0}'.format(account_id),
            self.authorization
        )

    def _update(self, data):
        self.access_token = data['access_token']
        self.expires_in = data['expires_in']
        self.expires_at = self.client.from_iso(data["expires_at"])
        self.token_type = data['token_type']
        self.refresh_token = data['refresh_token']
        self.refresh_expires = data['refresh_expires']
        self.refresh_expires_at = data['refresh_expires_at']
        self.account_id = data['account_id']
        self.client_id = data['client_id']
        self.internal_client = data['internal_client']
        self.client_service = data['client_service']
        self.app = data['app']
        self.in_app_id = data['in_app_id']
    
    async def schedule_token_refresh(self):
        self.token_timeout = (self.expires_at - datetime.datetime.utcnow()).total_seconds() - 300
        await asyncio.sleep(self.token_timeout)
        await self.do_refresh()

    async def do_refresh(self):
        log.debug('Refreshing session')

        if self.client.user.party is not None:
            await self.client.user.party._leave()

        data = await self.grant_refresh_token(self.refresh_token)
        self.launcher_access_token = data['access_token']
        await self.exchange_code('bearer {}'.format(self.launcher_access_token))

        log.debug('Refreshing xmpp session')
        await self.client.xmpp.close()
        await self.client.xmpp.run()

        await self.client._create_party()
        await self.schedule_token_refresh()



# -*- coding: utf-8 -*-

"""
MIT License

Copyright (c) 2019 Terbau

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import datetime
import aiohttp
import asyncio
import sys
import signal
import logging
import json

from .errors import EventError, PartyError, HTTPException
from .xmpp import XMPPClient
from .auth import Auth
from .http import HTTPClient
from .user import ClientUser, User
from .friend import Friend, PendingFriend
from .enums import PartyPrivacy
from .cache import Cache, WeakrefCache
from .party import ClientParty
from .stats import StatsV2
from .store import Store
from .news import BattleRoyaleNewsPost
from .playlist import Playlist

# logging.basicConfig(level=logging.DEBUG)
log = logging.getLogger(__name__) 

# all credit for this function goes to discord.py. Great task cancelling.
def _cancel_tasks(loop):
    task_retriever = asyncio.Task.all_tasks
    tasks = {t for t in task_retriever(loop=loop) if not t.done()}

    if not tasks:
        return

    log.info('Cleaning up after %d tasks.', len(tasks))
    for task in tasks:
        task.cancel()

    loop.run_until_complete(asyncio.gather(*tasks, loop=loop, return_exceptions=True))
    log.info('All tasks finished cancelling.')

    for task in tasks:
        if task.cancelled():
            continue
        if task.exception() is not None:
            loop.call_exception_handler({
                'message': 'Unhandled exception during Client.run shutdown.',
                'exception': task.exception(),
                'task': task
            })


def _cleanup_loop(loop):
    try:
        _cancel_tasks(loop)
        if sys.version_info >= (3, 6):
            loop.run_until_complete(loop.shutdown_asyncgens())
    finally:
        log.info('Closing the event loop.')
        loop.close()


class Client:
    """Represents the client connected to Fortnite and EpicGames' services.

    Parameters
    ----------
    email: Required[:class:`str`]
        The email of the account you want the client to log in as.
    password: Required[:class:`str`]
        The password of the account you want the client to log in as.
    two_factor_code: Optional[:class:`int`]
        The two-factor code to use when authenticating. If no two factor code is set
        and the passed account has 2fa enabled, you will be asked to enter it through
        console.
    loop: Optional[:class:`asyncio.AbstractEventLoop`]
        The event loop to use for asynchronous operations.
    status: :class:`str`
        The status you want the client to send with its presence to friends.
        Defaults to: ``Battle Royale Lobby - {party playercount} / {party max playercount}``
    platform: Optional[:class:`str`]
        The platform you want the client to display as its source. 
        Defaults to ``WIN``.
    net_cl: :class:`str`
        The current buildid used by the current Fortnite build. Named *netCL* in official logs.
        Defaults to the current buildid but doesn't get updated automatically. 
        
        .. warning::

            When a new buildid is pushed by EpicGames you must either wait for the library to get updated
            and then download the updated version with the correct buildid or you can get it yourself from
            the official logs and initialize client with it.

    default_party_config: Optional[:class:`dict`]
        The party configuration used when creating parties.
        Defaults to:

        .. code-block:: python3

            {
                'privacy': PartyPrivacy.PUBLIC,
                'joinability': 'OPEN',
                'max_size': 16,
                'sub_type': 'default',
                'type': 'default',
                'invite_ttl_seconds': 14400,
                'chat_enabled': True,
            }

    build: :class:`str`
        The build used by Fortnite. 
        Defaults to ``++Fortnite+Release-9.21-CL-6922310``

        .. note::

            The build is updated with every major version but is not that important to
            update as netCL.
    
    engine_build: :class:`str`
        The engine build used by Fortnite.
        Defaults to ``4.23.0-6922310+++Fortnite+Release-9.21``

        .. note::

            The build is updated with every major version but is not that important to
            update as netCL.
    
    launcher_token: :class:`str`
        The token used by EpicGames Launcher.
    fortnite_token: :class:`str`
        The token used by Fortnite.
    service_host: :class:`str`
        The host used by Fortnite's XMPP services.
    service_domain: :class:`str`
        The domain used by Fortnite's XMPP services.
    serivce_port: :class:`int`
        The port used by Fortnite's XMPP services.
    device_id: :class:`str`
        The hardware address of your computer as a 32char hex string.

    Attributes
    ----------
    user: :class:`ClientUser`
        The user the client is logged in as.
    loop: :class:`asyncio.AbstractEventLoop`
        The event loop that client implements.
    """

    def __init__(self, email, password, two_factor_code=None, loop=None, **kwargs):
        self.email = email
        self.password = password
        self.two_factor_code = two_factor_code
        self.loop = loop or asyncio.get_event_loop()

        self.status = kwargs.get('status', None)
        self.platform = kwargs.get('platform', 'WIN')
        self.net_cl = kwargs.get('net_cl', '7605985')
        self.party_build_id = '1:1:{0.net_cl}'.format(self)
        self.default_party_config = kwargs.get('default_party_config', {})
        self.build = kwargs.get('build', '++Fortnite+Release-10.10-CL-7955722')
        self.engine_build = kwargs.get('engine_build', '4.23.0-7955722+++Fortnite+Release-10.10')
        self.launcher_token = kwargs.get('launcher_token',
            'MzQ0NmNkNzI2OTRjNGE0NDg1ZDgxYjc3YWRiYjIxNDE6OTIwOWQ0YTVlMjVhNDU3ZmI5YjA3NDg5ZDMxM2I0MWE='
        )
        self.fortnite_token = kwargs.get('fortnite_token',
            'ZWM2ODRiOGM2ODdmNDc5ZmFkZWEzY2IyYWQ4M2Y1YzY6ZTFmMzFjMjExZjI4NDEzMTg2MjYyZDM3YTEzZmM4NGQ='
        )
        self.service_host = kwargs.get('xmpp_host', 'prod.ol.epicgames.com')
        self.service_domain = kwargs.get('xmpp_domain', 'xmpp-service-prod.ol.epicgames.com')
        self.service_port = kwargs.get('xmpp_port', 5222)
        self.device_id = kwargs.get('device_id', None)

        self.kill_other_sessions = True
        self.accept_eula = True

        self.http = HTTPClient(self, connector=kwargs.get('connector'))
        self.http.add_header('Accept-Language', 'en-EN')
        self.xmpp = None

        self._listeners = {}
        self._events = {}
        self._friends = Cache()
        self._pending_friends = Cache()
        self._users = Cache()
        self._presences = Cache()
        self.event_prefix = 'event'
        self._ready = asyncio.Event(loop=self.loop)
        self._refresh_task = None
        self._closed = False
        self._closing = False

        self.refresh_i = 0

        self.update_default_party_config(
            kwargs.get('default_party_config')
        )
        self._check_party_confirmation()
        
    @staticmethod
    def from_iso(iso):
        """:class:`str`: Converts an iso formatted string to a :class:`datetime.datetime` object
        
        Returns
        -------
        :class:`datetime.datetime`
        """
        if isinstance(iso, datetime.datetime):
            return iso
        return datetime.datetime.strptime(iso, '%Y-%m-%dT%H:%M:%S.%fZ')
    
    @staticmethod
    def to_iso(dt):
        """:class:`datetime.datetime`: Converts a :class:`datetime.datetime` object to an iso 
        formatted string

        Returns
        -------
        :class:`str`
        """
        return dt.strftime('%Y-%m-%dT%H:%M:%S.%fZ')

    @property
    def friends(self):
        """:class:`dict`: Mapping of current friends. {id (:class:`str`), :class:`Friend`}"""
        return self._friends._cache

    @property
    def pending_friends(self):
        """:class:`dict`: Mapping of currently pending friends. {id (:class:`str`), :class:`PendingFriend`}
        
        .. note::
        
            Pending friends can be both inbound (pending friend sent the request to the bot) or outgoing 
            (the bot sent the request to the pending friend).
        """
        return self._pending_friends._cache

    @property
    def presences(self):
        """:class:`dict`: Mapping of the last presence received from friends. {id (:class:`str`), :class:`Presence`}"""
        return self._presences._cache

    def update_default_party_config(self, config):
        if config is None:
            return

        if not isinstance(config, dict):
            raise PartyError('\'config\' must be a dictionary')

        _default_conf = {
            'privacy': PartyPrivacy.PUBLIC.value,
            'join_confirmation': False,
            'joinability': 'OPEN',
            'max_size': 16,
            'sub_type': 'default',
            'type': 'default',
            'invite_ttl_seconds': 14400,
            'chat_enabled': True,
        }

        try:
            self.default_party_config['privacy'] = self.default_party_config['privacy'].value
        except (KeyError, AttributeError):
            pass

        default_config = {**_default_conf, **self.default_party_config}
        self.default_party_config = {**default_config, **config}

    def _check_party_confirmation(self):
        try:
            getattr(self, '{0.event_prefix}_party_member_confirmation'.format(self))
            val = True
        except AttributeError:
            val = False

        self.update_default_party_config({'join_confirmation': val})

    def exc_handler(self, loop, ctx):
        log.debug('Exception was catched by asyncio exception handler: {}'.format(ctx['message']))

    def run(self):
        """This function starts the loop and then calls :meth:`start` for you.
        If you have passed an already running event loop to the client, you should start the client
        with :meth:`start`.

        .. warning::

            This function is blocking and should be the last function to run.

        Raises
        ------
        AuthException
            An error occured when attempting to log in.
        """
        loop = self.loop

        try:
            loop.add_signal_handler(signal.SIGINT, lambda: loop.stop())
            loop.add_signal_handler(signal.SIGTERM, lambda: loop.stop())
        except NotImplementedError:
            pass

        async def runner():
            try:
                await self.start()
            finally:
                if not self._closing:
                    await self.logout()

        asyncio.ensure_future(runner(), loop=loop)

        try:
            loop.run_forever()
        except KeyboardInterrupt:
            log.info('Terminating event loop.')
        finally:
            if not self._closing:
                log.info('Client not logged out when terminating loop. Logging out now.')
                loop.run_until_complete(self.logout())

            log.info('Cleaning up loop')
            _cleanup_loop(loop)
        
    async def start(self):
        """|coro|
        
        Starts the client and logs into the specified user.
        
        .. warning::

            This functions is blocking and everything after the line calling this function will never run!
            If you are using this function instead of :meth:`run` you should always call it after everything
            else. When the client is ready it will dispatch :meth:`event_ready`. 

        Raises
        ------
        AuthException
            An error occured when attempting to log in.
        """
        if self._closed:
            self.http.create_connection()
            self._closed = False

        await self._login()

        self._set_ready()
        self.dispatch_event('ready')

        self._refresh_task = self.loop.create_task(self.auth.schedule_token_refresh())
        try:
            await self._refresh_task
        except asyncio.CancelledError:
            pass
        
    async def _login(self):
        self.auth = Auth(self)
        await self.auth.authenticate()

        data = await self.http.get_profile(self.auth.account_id)
        self.user = ClientUser(self, data)

        if self.kill_other_sessions:
            await self.auth.kill_other_sessions()
            log.debug('Other sessions killed')

        if self.accept_eula:
            await self.auth.accept_eula(self.auth.account_id)
            log.debug('EULA accepted')

        await self.initialize_friends()

        self.xmpp = XMPPClient(self)
        await self.xmpp.run()
        log.debug('Connected to XMPP')

        await self.initialize_party()
        log.debug('Party created')

    async def logout(self):
        """|coro|
        
        Logs the user out and closes running services.

        Raises
        ------
        HTTPException
            An error occured while logging out.
        """
        self._closing = True
        if self._refresh_task is not None and not self._refresh_task.cancelled():
            self._refresh_task.cancel()

        try:
            if self.user.party is not None:
                await self.user.party._leave()
        except:
            pass

        try:
            await self.xmpp.close()
        except:
            pass

        try:
            await self.auth.kill_token(self.auth.access_token)
        except:
            pass

        self._friends.clear()
        self._pending_friends.clear()
        self._users.clear()
        self._presences.clear()
        self._ready.clear()
        
        await self.http.close()
        self._closed = True
        self._closing = False
        log.debug('Successfully logged out')

    def _set_ready(self):
        self._ready.set()

    def is_ready(self):
        """Specifies if the internal state of the client is ready.
        
        Returns
        -------
        :class:`bool`
            ``True`` if the internal state is ready else ``False``
        """
        return self._ready.is_set()

    async def wait_until_ready(self):
        """|coro|
        
        Waits until the internal state of the client is ready.
        """
        await self._ready.wait()

    async def initialize_party(self):
        data = await self.http.party_lookup_user(self.user.id)
        if len(data['current']) > 0:
            party = ClientParty(self, data['current'][0])
            await party._leave()
            log.debug('Left old party')
        await self._create_party()

    async def fetch_profile_by_display_name(self, display_name, cache=True, raw=False):
        """|coro|
        
        Fetches a profile from the passed display name
        
        Parameters
        ----------
        display_name: :class:`str`
            The display name of the user you want to fetch the profile for.
        cache: Optional[:class:`bool`]
            If set to True it will try to get the profile from the friends or user cache.
            *Defaults to ``True``*

            .. note::

                Setting this parameter to False will make it an api call.

        raw: Optional[:class:`bool`]
            If set to True it will return the data as you would get it from the api request.
            *Defaults to ``False``*

            .. note::

                Setting raw to True does not work with cache set to True.

        Raises
        ------
        HTTPException
            An error occured while requesting the user.

        Returns
        -------
        :class:`User`
            The user requested. If not found it will return ``None``  
        """
        if cache:
            for u in self._users.values():
                try:
                    if u.display_name.lower() == display_name.lower():
                        return u
                except AttributeError:
                    pass
        try:
            res = await self.http.get_profile_by_display_name(display_name)
        except HTTPException as exc:
            if exc.message_code != 'errors.com.epicgames.account.account_not_found':
                raise HTTPException(exc.response, exc.raw)
            return None

        if raw:
            return res
        return self.store_user(res)

    async def fetch_profile(self, user, cache=True, raw=False):
        """|coro|
        
        Fetches a single profile by the given id/displayname

        Parameters
        ----------
        user: :class:`str`
            Id or display name
        cache: Optional[:class:`bool`]
            If set to True it will try to get the profile from the friends or user cache
            and fall back to an api request if not found.
            *Defaults to ``True``*

            .. note::

                Setting this parameter to False will make it an api call.

        raw: Optional[:class:`bool`]
            If set to True it will return the data as you would get it from the api request.
            *Defaults to ``False``*

            .. note::

                Setting raw to True does not work with cache set to True.

        Raises
        ------
        HTTPException
            An error occured while requesting the user.

        Returns
        -------
        :class:`User`
            The user requested. If not found it will return ``None``
        """
        try:
            return (await self.fetch_profiles((user,), cache=cache, raw=raw))[0]
        except IndexError:
            return None

    async def fetch_profiles(self, users, cache=True, raw=False):
        """|coro|
        
        Fetches multiple profiles at once by the given ids/displaynames

        Parameters
        ----------
        users: List[:class:`str`]
            A list/tuple containing ids/displaynames.
        cache: Optional[:class:`bool`]
            If set to True it will try to get the profiles from the friends or user cache
            and fall back to an api request if not found.
            *Defaults to ``True``*

            .. note::

                Setting this parameter to False will make it an api call.

        raw: Optional[:class:`bool`]
            If set to True it will return the data as you would get it from the api request.
            *Defaults to ``False``*

            .. note::

                Setting raw to True does not work with cache set to True.

        Raises
        ------
        HTTPException
            An error occured while requesting user information.

        Returns
        -------
        List[:class:`User`]
            Users requested. Only users that are found gets returned.
        """
        if len(users) == 0:
            return []

        profiles = []
        new = []
        tasks = []
        for elem in users:
            if self.is_display_name(elem):
                for u in self._users.values():
                    try:
                        if u.display_name.lower() == elem.lower():
                            profiles.append(u)
                            continue
                    except AttributeError:
                        pass

                task = self.http.get_profile_by_display_name(elem)
                tasks.append(task)
            else:
                if cache:
                    p = self.get_user(elem)
                    if p:
                        if raw:
                            profiles.append(p.get_raw())
                        else:
                            profiles.append(p)
                        continue
                new.append(elem)

        if len(tasks) > 0:
            done, _ = await asyncio.wait(tasks)
            for f in done:
                id = (f.result())['id']
                new.append(id)

        chunk_tasks = []
        chunks = [new[i:i + 100] for i in range(0, len(new), 100)]
        for chunk in chunks:
            task = self.http.get_profiles(chunk)
            chunk_tasks.append(task)
        
        if len(chunks) > 0:
            d, _ = await asyncio.wait(chunk_tasks)
            for results in d:
                for result in results.result():
                    if raw:
                        profiles.append(result)
                    else:
                        u = self.store_user(result)
                        profiles.append(u)
        return profiles

    async def initialize_friends(self):
        raw_friends = await self.http.get_friends(include_pending=True)
        ids = [r['accountId'] for r in raw_friends]
        friends = await self.fetch_profiles(ids, raw=True)
        
        m = {}
        for friend in friends:
            m[friend['id']] = friend

        for friend in raw_friends:
            if friend['status'] == 'PENDING':
                try:
                    f = PendingFriend(self, {**friend, **m[friend['accountId']]})
                    self._pending_friends.set(f.id, f)
                except KeyError:
                    continue

            elif friend['status'] == 'ACCEPTED':
                try:
                    f = Friend(self, {**friend, **m[friend['accountId']]})
                    self._friends.set(f.id, f)
                except KeyError:
                    continue

    def store_user(self, data):
        try:
            return self._users.get(data['id'], silent=False)
        except KeyError:
            u = User(self, data)
            self._users.set(u.id, u)
            return u

    def get_user(self, id):
        """Tries to get a user from the user cache by the given id.

        Parameters
        ----------
        id: :class:`str`
            The id of the user.

        Returns
        -------
        :class:`User`
            :class:`User` if found, else ``None``
        """
        user = self._users.get(id)
        if user is None:
            friend = self.get_friend(id)
            if friend is not None:
                user = User(self, friend.get_raw())
                self._users.set(user.id, user)
        return user

    def store_friend(self, data):
        try:
            return self._friends.get(data['accountId'], silent=False)
        except KeyError:
            f = Friend(self, data)
            self._friends.set(f.id, f)
            return f

    def get_friend(self, id):
        """Tries to get a friend from the friend cache by the given id.

        Parameters
        ----------
        id: :class:`str`
            The id of the friend.

        Returns
        -------
        :class:`Friend`
            :class:`Friend` if found, else ``None``
        """
        return self._friends.get(id)

    def get_presence(self, id):
        """Tries to get the latest received presence from the presence cache.

        Parameters
        ----------
        id: :class:`str`
            The id of the friend you want the last presence of.

        Returns
        -------
        :class:`Presence`
            :class:`Presence` if found, else ``None``
        """
        return self._presences.get(id)
    
    def get_pending_friend(self, id):
        """Tries to get a pending friend from the pending friend cache by the given id.

        Parameters
        ----------
        id: :class:`str`
            The id of the pending friend.

        Returns
        -------
        :class:`PendingFriend`: 
            :class:`PendingFriend` if found, else ``None``
        """
        return self._pending_friends.get(id)

    def has_friend(self, id):
        """Checks if the client is friends with the given id.

        Parameters
        ----------
        id: :class:`str`
            The id of the user you want to check.

        Returns
        -------
        :class:`bool`
            ``True`` if user is friends with the client else ``False``
        """
        return self.get_friend(id) is not None
    
    def is_pending(self, id):
        """Checks if the given id is a pending friend of the client.

        Parameters
        ----------
        id: :class:`str`
            The id of the user you want to check.

        Returns
        -------
        :class:`bool`
            ``True`` if user is a pending friend else ``False``
        """
        return self.get_pending_friend(id)

    async def get_blocklist(self):
        """|coro|
        
        Retrieves the blocklist with an api call.

        Raises
        ------
        HTTPException
            An error occured while fetching blocklist.

        Returns
        -------
        List[:class:`str`]
            List of ids
        """
        return (await self.http.get_friends_blocklist())['blockedUsers']

    async def block_user(self, id):
        """|coro|
        
        Blocks a user by a given id.

        Parameters
        ----------
        id: :class:`str`
            The id of the user you want to block.

        Raises
        ------
        HTTPException
            Something went wrong when trying to block this user.
        """
        await self.http.block_user(id)

    async def unblock_user(self, id):
        """|coro|
        
        Unblocks a user by a given id.

        Parameters
        ----------
        id: :class:`str`
            The id of the user you want to unblock

        Raises
        ------
        HTTPException
            Something went wrong when trying to unblock this user.
        """
        await self.http.unblock_user(id)

    def is_id(self, value):
        """Simple function that finds out if a :class:`str` is a valid id

        Parameters
        ----------
        value: :class:`str`
            The string you want to check.

        Returns
        -------
        :class:`bool`
            ``True`` if string is valid else ``False``
        """
        return isinstance(value, str) and len(value) > 16

    def is_display_name(self, val):
        """Simple function that finds out if a :class:`str` is a valid displayname

        Parameters
        ----------
        value: :class:`str`
            The string you want to check.

        Returns
        -------
        :class:`bool`
            ``True`` if string is valid else ``False``
        """
        return isinstance(val, str) and 3 <= len(val) <= 16  

    async def add_friend(self, id):
        """|coro|
        
        Sends a friend request or accepts a request if found.

        Raises
        ------
        HTTPException
            An error occured while requesting to add this friend.

        Parameters
        ----------
        id: :class:`str`
            The id of the user you want to add/accept.
        """
        await self.http.add_friend(id)
    
    async def remove_friend(self, id):
        """|coro|
        
        Removes a friend by the given id.

        Parameters
        ----------
        id: :class:`str`
            The id of the friend you want to remove.

        Raises
        ------
        HTTPException
            Something went wrong when trying to remove this friend.
        """
        await self.http.remove_friend(id)

    def dispatch_event(self, event, *args, **kwargs):
        method = '{0.event_prefix}_{1}'.format(self, event)

        listeners = self._listeners.get(event)
        if listeners:
            removed = []
            for i, (future, check) in enumerate(listeners):
                if future.cancelled():
                    removed.append(i)
                    continue
                
                try:
                    result = check(*args)
                except Exception as e:
                    future.set_exception(e)
                    removed.append(i)
                else:
                    if result:
                        if len(args) == 0:
                            future.set_result(None)
                        elif len(args) == 1:
                            future.set_result(args[0])
                        else:
                            future.set_result(args)
                        removed.append(i)

            if len(removed) == len(listeners):
                self._listeners.pop(event)
            else:
                for idx in reversed(removed):
                    del listeners[idx]
        
        try:
            coro = getattr(self, method)
        except AttributeError as e:
            pass
        else:
            asyncio.ensure_future(coro(*args, **kwargs), loop=self.loop)

        if event in self._events.keys():
            for coro in self._events[event]:
                asyncio.ensure_future(coro(*args, **kwargs), loop=self.loop)

    def wait_for(self, event, *, check=None, timeout=None):
        """|coro|
        
        Waits for an event to be dispatch.
        
        In case the event returns more than one arguments, a tuple is passed containing 
        the arguments.
        
        Examples
        --------
        This example waits for the author of a :class:`FriendMessage` to say hello.: ::

            @client.event
            async def event_friend_message(message):
                await message.reply('Say hello!')

                def check_function(m):
                    return m.author.id == message.author.id
                
                msg = await client.wait_for('message', check=check_function, timeout=60)
                await msg.reply('Hello {0.author.display_name}!'.format(msg))

        This example waits for the the leader of a party to promote the bot after joining
        and then sets a new custom key: ::

            @client.event
            async def event_party_member_join(member):

                # checks if the member that joined is the UserClient
                if member.id != client.user.id:
                    return

                def check(m):
                    return m.id == client.user.id
                
                try:
                    await client.wait_for('party_member_promote', check=check, timeout=120)
                except asyncio.TimeoutError:
                    await member.party.send('You took too long to promote me!')
                
                await member.party.set_custom_key('my_custom_key_123')

        Parameters
        ----------
        event: :class:`str`
            The name of the event. 

            .. note::

                The name of the event must be **without** the ``event_`` prefix.

                Wrong = ``event_friend_message``.
                Correct = ``friend_message``.

        check: Optional[Callable function]
            A predicate to check what to wait for.
            *Defaults to a predicate that always returns ``True``. This means it will return the first
            result.*

        timeout: :class:`int`
            How many seconds to wait for before asyncio.TimeoutError is raised.
            *Defaults to ``None`` which means it will wait forever.*

        Raises
        ------
        asyncio.TimeoutError
            No event was retrieved in the time you specified.

        Returns
        -------
        Any
            Returns arguments based on the event you are waiting for. An event might return
            no arguments, one argument or a tuple of arguments. Check the
            :ref:`event reference <fortnitepy-events-api> for more information about the returning
            arguments.`
        """
        future = self.loop.create_future()
        if check is None:
            def _check(*args):
                return True
            check = _check
        
        ev = (event.lower()).replace('{0.event_prefix}_'.format(self), '')
        try:
            listeners = self._listeners[ev]
        except KeyError:
            listeners = []
            self._listeners[ev] = listeners
        
        listeners.append((future, check))
        return asyncio.wait_for(future, timeout, loop=self.loop)

    def add_event_handler(self, event, coro):
        if not asyncio.iscoroutinefunction(coro):
            raise TypeError('event registered must be a coroutine function')

        if event not in self._events.keys():
            self._events[event] = []
        self._events[event].append(coro)
    
    def remove_event_handler(self, event, coro):
        if event not in self._events.keys():
            return
        
        try:
            self._events[event].remove(coro)
        except ValueError:
            pass

    def event(self, coro):
        """A decorator to register an event.
        
        .. note::

            You do not need to decorate events in a subclass of :class:`Client`.

        Usage: ::

            @client.event
            async def event_friend_message(message):
                await message.reply('Thanks for your message!')

        Raises
        ------
        TypeError
            The name of the function does is not prefixed with ``event_``
        TypeError
            The decorated function is not a coroutine.
        """
        if not coro.__name__.startswith('event_'):
            raise TypeError('event function names must follow this syntax: "event_<event>"')
        if not asyncio.iscoroutinefunction(coro):
            raise TypeError('the decorated function must be a coroutine')

        name = '_'.join(coro.__name__.split('_')[1:])
        self.add_event_handler(name, coro)
        log.debug('{} has been registered as an event'.format(coro.__name__))
        return coro

    async def fetch_br_stats(self, user_id, start_time=None, end_time=None):
        """|coro|
        
        Gets Battle Royale stats the specified user.

        Parameters
        ----------
        user_id: :class:`str`
            The id of the user you want to fetch stats for.
        start_time: Optional[Union[:class:`int`, :class:`datetime.datetime`]]
            The UTC start time of the time period to get stats from.
            *Must be seconds since epoch or :class:`datetime.datetime`*
            *Defaults to None*
        end_time: Optional[Union[:class:`int`, :class:`datetime.datetime`]]
            The UTC end time of the time period to get stats from.
            *Must be seconds since epoch or :class:`datetime.datetime`*
            *Defaults to None*

        Raises
        ------
        HTTPException
            An error occured while requesting.
        
        Returns
        -------
        :class:`StatsV2`
            An object representing the stats for this user.
        """
        epoch = datetime.datetime.utcfromtimestamp(0)
        if isinstance(start_time, datetime.datetime):
            start_time = (start_time - epoch).total_seconds()

        if isinstance(end_time, datetime.datetime):
            end_time = (end_time - epoch).total_seconds()

        data = await self.http.get_br_stats_v2(user_id, start_time=start_time, end_time=end_time)
        return StatsV2(data)

    async def fetch_multiple_br_stats(self, user_ids, stats, start_time=None, end_time=None):
        """|coro|
        
        Gets Battle Royale stats for multiple users at the same time.
        
        .. note::
            
            This function is not the same as doing :meth:`fetch_br_stats` for multiple users.
            The expected return for this function would not be all the stats for the specified
            users but rather the stats you specified.

        Example usage: ::

            async def stat_function():
                stats = [
                    fortnitepy.StatsV2.create_stat('placetop1', fortnitepy.V2Inputs.KEYBOARDANDMOUSE, 'defaultsolo'),
                    fortnitepy.StatsV2.create_stat('kills', fortnitepy.V2Inputs.KEYBOARDANDMOUSE, 'defaultsolo'),
                    fortnitepy.StatsV2.create_stat('matchesplayed', fortnitepy.V2Inputs.KEYBOARDANDMOUSE, 'defaultsolo')
                ]

                # get the profiles and create a list of their ids.
                profiles = await self.fetch_profiles(['Ninja', 'Dark', 'DrLupo'])
                user_ids = [u.id for u in profiles]

                data = await self.fetch_multiple_br_stats(user_ids=user_ids, stats=stats)
                for id, res in data.items():
                    print('ID: {0} | Stats: {1}'.format(id, res.stats))
            
            # expected output (ofc the values would be updated):
            # ID: 463ca9d604524ce38071f512baa9cd70 | Stats: {'keyboardmouse': {'defaultsolo': {'wins': 759, 'kills': 28093, 'matchesplayed': 6438}}}
            # ID: 3900c5958e4b4553907b2b32e86e03f8 | Stats: {'keyboardmouse': {'defaultsolo': {'wins': 1763, 'kills': 41375, 'matchesplayed': 7944}}}
            # ID: 4735ce9132924caf8a5b17789b40f79c | Stats: {'keyboardmouse': {'defaultsolo': {'wins': 1888, 'kills': 40784, 'matchesplayed': 5775}}}

        Parameters
        ----------
        user_ids: List[:class:`str`]
            A list of ids you are requesting the stats for.
        stats: List[:class:`str`]
            A list of stats to get for the users. Use :meth:`StatsV2.create_stat` to create the stats.

            Example: ::

                [
                    fortnitepy.StatsV2.create_stat('placetop1', fortnitepy.V2Inputs.KEYBOARDANDMOUSE, 'defaultsolo'),
                    fortnitepy.StatsV2.create_stat('kills', fortnitepy.V2Inputs.KEYBOARDANDMOUSE, 'defaultsolo'),
                    fortnitepy.StatsV2.create_stat('matchesplayed', fortnitepy.V2Inputs.KEYBOARDANDMOUSE, 'defaultsolo')
                ]

        start_time: Optional[Union[:class:`int`, :class:`datetime.datetime`]]
            The UTC start time of the time period to get stats from.
            *Must be seconds since epoch or :class:`datetime.datetime`*
            *Defaults to None*
        end_time: Optional[Union[:class:`int`, :class:`datetime.datetime`]]
            The UTC end time of the time period to get stats from.
            *Must be seconds since epoch or :class:`datetime.datetime`*
            *Defaults to None*

        Raises
        ------
        HTTPException
            An error occured while requesting.

        Returns
        -------
        Dict[id: :class:`StatsV2`]
            A mapping where :class:`StatsV2` is bound to its owners id.
        """
        epoch = datetime.datetime.utcfromtimestamp(0)
        if isinstance(start_time, datetime.datetime):
            start_time = (start_time - epoch).total_seconds()

        if isinstance(end_time, datetime.datetime):
            end_time = (end_time - epoch).total_seconds()
        
        data = await self.http.get_multiple_br_stats_v2(user_ids, stats)

        res = {}
        for udata in data:
            res[udata['accountId']] = StatsV2(udata)
        return res

    async def fetch_leaderboard(self, stat):
        """|coro|
        
        Fetches the leaderboard for a stat.
        
        .. warning::
        
            For some weird reason, the only valid stat you can pass is
            one with ``placetop1`` (``wins`` is also accepted).
            
        Example usage: ::

            async def get_leaderboard():
                stat = fortnitepy.StatsV2.create_stat(
                    'wins',
                    fortnitepy.V2Inputs.KEYBOARDANDMOUSE,
                    'defaultsquad'
                )

                data = await client.fetch_leaderboard(stat)

                for placement, entry in enumerate(data):
                    print('[{0}] Id: {1} | Wins: {2}'.format(
                        placement, entry['account'], entry['value']))

        Parameters
        ----------
        stat: :class:`str`
            The stat you are requesting the leaderboard entries for. You can use 
            :meth:`StatsV2.create_stat` to create this string.
            
        Raises
        ------
        ValueError
            You passed an invalid/non-accepted stat argument. 
        HTTPException
            An error occured when requesting.

        Returns
        -------
        List[Dict['account': :class:`str`: id, 'value': :class:`int`: The stat value.]]
            List of dictionaries containing entry data. Example return: ::

                {
                    'account': '4480a7397f824fe4b407077fb9397fbb',
                    'value': 5082
                }
        """
        data = await self.http.get_br_leaderboard_v2(stat)

        if len(data['entries']) == 0:
            raise ValueError('{0} is not a valid stat'.format(stat))
        
        return data['entries']

    async def _create_party(self, config=None):
        if isinstance(config, dict):
            cf = {**self.default_party_config, **config}
        else:
            cf = self.default_party_config

        while True:
            try:
                data = await self.http.party_create(cf)
                # print(json.dumps(data, indent=2))
                break
            except HTTPException as exc:
                if exc.message_code != 'errors.com.epicgames.social.party.user_has_party':
                    raise HTTPException(exc.response, exc.raw)

                data = await self.http.party_lookup_user(self.user.id)
                await self.http.party_leave(data['current'][0]['id'])

        config = {**cf, **data['config']}
        party = ClientParty(self, data)
        await party._update_members(data['members'])
        asyncio.ensure_future(party.join_chat(), loop=self.loop)
        self.user.set_party(party)

        def check(m):
            return m.id == self.user.id

        try:
            await self.wait_for('party_member_join', check=check, timeout=3)
        except asyncio.TimeoutError:
            await party._leave()
            return await self._create_party()

        await party.set_privacy(config['privacy'])
        return party

    async def join_to_party(self, party_id, party=None):
        if party is None:
            party_data = await self.http.party_lookup(party_id)
            party = ClientParty(self, party_data)
            await party._update_members(party_data['members'])

        await self.user.party._leave()
        self.user.set_party(party)

        await self.http.party_join_request(party_id)
        await party._update_members_meta()
        asyncio.ensure_future(self.user.party.join_chat(), loop=self.loop)

    async def set_status(self, status):
        """|coro|
        
        Sends and sets the status. This status message will override all other
        presence statuses including party presence status.

        Parameters
        ----------
        status: :class:`str`
            The status you want to set.

        Raises
        ------
        TypeError
            The status you tried to set were not a str.
        """
        if not isinstance(status, str):
            raise TypeError('status must be a str')
        
        self.status = status
        await self.xmpp.send_presence(status=status)

    async def send_status(self, status, to=None):
        """|coro|
        
        Sends this status to all or one single friend.
        
        Parameters
        ----------
        status: Union[:class:`str`, :class:`dict`]
            The status message in :class:`str` or full status in :class:`dict`.
        to: Optional[:class:`aioxmpp.JID`]
            The JID of the user that should receive this status.
            *Defaults to None which means it will send to all friends.*

        Raises
        ------
        TypeError
            Status was an invalid type.
        """
        await self.xmpp.send_presence(status=status, to=to)

    async def fetch_lightswitch_status(self, service_id='Fortnite'):
        """|coro|
        
        Fetches the lightswitch status of an epicgames service.

        Parameters
        ----------
        service_id: Optional[:class:`str`]
            The service id to check status for.
            *Defaults to ``Fortnite``.

        Raises
        ------
        ValueError
            The returned data was empty. Most likely because service_id is not valid.
        HTTPException
            An error occured when requesting.

        Returns
        -------
        :class:`bool`
            ``True`` if service is up else ``False``
        """
        status = await self.http.get_lightswitch_status(service_id=service_id)
        if len(status) == 0:
            raise ValueError('emtpy lightswitch response')
        return True if status[0].get('status') == 'UP' else False

    async def fetch_item_shop(self):
        """|coro|
        
        Fetches the current item shop.

        Example: ::

            # fetches all CIDs (character ids) of of the current item shop.
            async def get_current_item_shop_cids():
                store = await client.fetch_item_shop()

                cids = []
                for item in store.featured_items + store.daily_items:
                    for grant in item.grants:
                        if grant['type'] == 'AthenaCharacter':
                            cids.append(grant['asset'])

                return cids

        Raises
        ------
        HTTPException
            An error occured when requesting.

        Returns
        -------
        :class:`Store`
            Object representing the data from the current item shop.
        """
        data = await self.http.get_store_catalog()
        return Store(self, data)

    async def fetch_br_news(self):
        """|coro|
        
        Fetches news for the Battle Royale gamemode.

        Raises
        ------
        HTTPException
            An error occured when requesting.

        Returns
        -------
        :class:`list`
            List[:class:`BattleRoyaleNewsPost`]
        """
        raw = await self.http.get_fortnite_news()
        data = json.loads(raw.encode('utf-8'))

        res = []
        msg = data['battleroyalenews']['news'].get('message')
        if msg is not None:
            res.append(BattleRoyaleNewsPost(msg))
        else:
            msgs = data['battleroyalenews']['news']['messages']
            for msg in msgs:
                res.append(BattleRoyaleNewsPost(msg))
        return res

    async def fetch_br_playlists(self):
        """|coro|
        
        Fetches all playlists registered on Fortnite. This includes all previous gamemodes
        that is no longer active.
        
        Raises
        ------
        HTTPException
            An error occured while requesting.
        
        Returns
        -------
        List[:class:`Playlist`]
            List containing all playlists registered on Fortnite.
        """
        data = await self.http.get_fortnite_content()

        raw = data['playlistinformation']['playlist_info']['playlists']
        playlists = []
        for playlist in raw:
            try:
                p = Playlist(playlist)
                playlists.append(p)
            except KeyError:
                pass
        return playlists

    async def fetch_active_ltms(self, region):
        """|coro|
        
        Fetches active LTMs for a specific region.
        
        Parameters
        ----------
        :class:`Regions`
            The region to request active LTMs for.
            
        Raises
        ------
        HTTPException
            An error occured while requesting.
            
        Returns
        -------
        List[:class:`str`]
            List of internal playlist names.
        """
        data = await self.http.get_fortnite_timeline()

        states = data['channels']['client-matchmaking']['states']
        region_data = states[len(states) - 1]['state']['region'][region.value]
        return region_data.get('eventFlagsForcedOn', [])



# -*- coding: utf-8 -*-

"""
MIT License

Copyright (c) 2019 Terbau

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import datetime
import logging
import aioxmpp

log = logging.getLogger(__name__)


class UserBase:
    __slots__ = ('client', '_display_name', '_id', '_external_auths')

    def __init__(self, client, data, **kwargs):
        self.client = client
        if data:
            self._update(data)

    @property
    def display_name(self):
        """:class:`str`: The users displayname"""
        return self._display_name
    
    @property
    def id(self):
        """:class:`str`: The users id"""
        return self._id
    
    @property
    def external_auths(self):
        """:class:`list`: List containing information about external auths.
        Might be empty if the user does not have any external auths"""
        return self._external_auths

    @property
    def jid(self):
        """:class:`aioxmpp.JID`: The JID of the user."""
        return aioxmpp.JID.fromstr('{0.id}@{0.client.service_host}'.format(self))

    def _update(self, data):
        self._display_name = data.get('displayName', data.get('account_dn'))
        self._external_auths = data.get('external_auths', [])

        try:
            self._id = data['id']
        except KeyError:
            self._id = data.get('accountId', data.get('account_id'))

    def get_raw(self):
        return {
            'displayName': self.display_name,
            'id': self.id,
            'externalAuths': self.external_auths
        }


class ClientUser(UserBase):
    """Represents the user the client is connected to.
    
    Attributes
    ----------
    client: :class:`Client`
        The client.
    age_group: :class:`str`
        The age group of the user.
    can_update_display_name: :class:`bool`
        ``True`` if the user can update it's displayname else ``False``
    country: :class:`str`
        The country the user wasregistered in.
    email: :class:`str`
        The email of the user.
    failed_login_attempts: :class:`str`
        Failed login attempts
    headless: :class:`str`
        COME BACK TO THIS
    last_login: :class:`datetime.datetime`
        UTC time of the last login of the user.
    name: :class:`str`
        First name of the user.
    first_name: :class:`str`
        First name of the user. Alias for name.
    last_name: :class:`str`
        Last name of the user.
    full_name: :class:`str`
        Full name of the user.
    number_of_display_name_changes: :class:`int`
        Amount of displayname changes.
    preferred_language: :class:`str`
        Users preferred language.
    tfa_enabled: :class:`bool`
        ``True`` if the user has two-factor authentication enabled.
    """

    def __init__(self, client, data, **kwargs):
        super().__init__(client, data)
        self._party = None
        self._update(data)

    @property
    def first_name(self):
        return self.name

    @property
    def full_name(self):
        return '{} {}'.format(self.name, self.last_name)

    @property
    def party(self):
        """:class:`Party`: The users party."""
        return self._party

    @property
    def jid(self):
        """:class:`aioxmpp.JID`: The JID of the client. Includes the 
        resource part.
        """
        return self.client.xmpp.xmpp_client.local_jid

    def _update(self, data):
        super()._update(data)
        self.age_group = data['ageGroup']
        self.can_update_display_name = bool(data['canUpdateDisplayName'])
        self.country = data['country']
        self.email = data['email']
        self.failed_login_attempts = data['failedLoginAttempts']
        self.headless = data['headless']
        self.last_login = self.client.from_iso(data['lastLogin'])
        self.name = data['name']
        self.last_name = data['lastName']
        self.number_of_display_name_changes = int(data['numberOfDisplayNameChanges'])
        self.preferred_language = data['preferredLanguage']
        self.tfa_enabled = bool(data['tfaEnabled'])

    def set_party(self, party):
        self._party = party

    def remove_party(self):
        self._party = None


class User(UserBase):

    __slots__ = UserBase.__slots__
    
    def __init__(self, client, data, **kwargs):
        super().__init__(client, data)



# -*- coding: utf-8 -*-

"""
MIT License

Copyright (c) 2019 Terbau

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from .user import UserBase
from .errors import PartyError, PartyPermissionError


class FriendBase(UserBase):

    __slots__ = UserBase.__slots__ + \
                ('_status', '_direction', '_favorite', '_created_at')

    def __init__(self, client, data):
        super().__init__(client, data)
        self._update(data)
        
    def _update(self, data):
        super()._update(data)
        self._status = data['status']
        self._direction = data['direction']
        self._favorite = bool(data['favorite'])
        self._created_at = self.client.from_iso(data['created'])

    @property
    def display_name(self):
        """:class:`str`: The friends' displayname"""
        return self._display_name
    
    @property
    def id(self):
        """:class:`str`: The friends' id"""
        return self._id
    
    @property
    def external_auths(self):
        """:class:`list`: List containing information about external auths.
        Might be empty if the friend does not have any external auths"""
        return self._external_auths

    @property
    def jid(self):
        """:class:`aioxmpp.JID`: The jid of the friend."""
        return super().jid

    @property
    def status(self):
        """:class:`str`: The friends status to the client. E.g. if the friend
        is friends with the bot it will be ``ACCEPTED``.
        
        .. warning::
        
            This is not the same as status from presence!
        
        """
        return self._status
    
    @property
    def direction(self):
        """:class:`str`: The direction of the friendship. ``INBOUND`` if the friend 
        added :class:`ClientUser` else ``OUTGOING``.
        """
        return self._direction

    @property
    def inbound(self):
        """:class:`bool`: ``True`` if this friend was the one to send the friend request else ``False``."""
        return self._direction == 'INBOUND'

    @property
    def outgoing(self):
        """:class:`bool`: ``True`` if the bot was the one to send the friend request else ``False``."""
        return self._direction == 'OUTGOING'

    @property
    def favorite(self):
        """:class:`bool`: ``True`` if the friend is favorited by :class:`ClientUser`
        else ``False``.
        """
        return self._favorite

    @property
    def created_at(self):
        """:class:`datetime.datetime`: The UTC time of when the friendship was created."""
        return self._created_at

    async def block(self):
        """|coro|
        
        Blocks this friend.

        Raises
        ------
        HTTPException
            Something went wrong when trying to block this user.
        """
        await self.client.http.block_user(self.id)

    def get_raw(self):
        return {
            **(super().get_raw()),
            'status': self.status,
            'direction': self.direction,
            'favorite': self.favorite,
            'created': self.created_at
        }


class Friend(FriendBase):
    """Represents a friend on Fortnite"""

    __slots__ = FriendBase.__slots__

    def __init__(self, client, data):
        super().__init__(client, data)

    @property
    def display_name(self):
        """:class:`str`: The friends displayname"""
        return self._display_name
    
    @property
    def id(self):
        """:class:`str`: The friends id"""
        return self._id
    
    @property
    def external_auths(self):
        """:class:`list`: List containing information about external auths.
        Might be empty if the friend does not have any external auths
        """
        return self._external_auths

    @property
    def last_presence(self):
        """:class:`Presence`: The last presence retrieved by the
        friend. Might be ``None`` if no presence has been 
        received by this friend yet.
        """
        return self.client.get_presence(self.id)

    @property
    def is_online(self):
        """:class:`bool`: ``True`` if this friend is currently online on 
        Fortnite else ``False``.
        """
        pres = self.client.get_presence(self.id)
        if pres is None:
            return False
        return pres.is_available

    async def remove(self):
        """|coro|
        
        Removes the friend from your friendlist.

        Raises
        ------
        HTTPException
            Something went wrong when trying to remove this friend.
        """
        await self.client.http.remove_friend(self.id)

    async def send(self, content):
        """|coro|
        
        Sends a :class:`FriendMessage` to this friend.

        Parameters
        ----------
        content: :class:`str`
            The content of the message.
        """
        await self.client.xmpp.send_friend_message(self.jid, content)

    async def join_party(self):
        """|coro|
        
        Attempts to join this friends' party.
        
        Raises
        ------
        PartyError
            Party was not found.
        PartyPermissionError
            The party you attempted to join was private.
        HTTPException
            Something else went wrong when trying to join the party.
        """
        _pre = self.last_presence
        if _pre is None:
            raise PartyError('Could not join party. Reason: Party not found')
        
        if _pre.party.is_private:
            raise PartyPermissionError('Could not join party. Reason: Party is private')
        
        await _pre.party.join()


class PendingFriend(FriendBase):
    """Represents a pending friend from Fortnite."""

    __slots__ = FriendBase.__slots__

    def __init__(self, client, data):
        super().__init__(client, data)

    @property
    def created_at(self):
        """:class:`datetime.datetime`: The UTC time of when the request was created"""
        return self._created_at

    async def accept(self):
        """|coro|
        
        Accepts this users' friend request.

        Raises
        ------
        HTTPException
            Something went wrong when trying to accept this request.
        """
        await self.client.http.add_friend(self.id)

    async def decline(self):
        """|coro|
        
        Declines this users' friend request.

        Raises
        ------
        HTTPException
            Something went wrong when trying to decline this request.
        """
        await self.client.http.remove_friend(self.id)


# NOT IMPLEMENTED
# class BlockedFriend(FriendBase):
#     """Represents a blocked friend from Fortnite"""
#     def __init__(self, client, data):
#         super().__init__(client, data)

#     @property
#     def created_at(self):
#         """:class:`datetime.datetime`: The time of when the user was blocked"""
#         return self.created_at

#     async def unblock(self):
#         """|coro|
        
#         Unblocks this friend."""
#         await self.client.http.unblock_user(self.id)

    


# -*- coding: utf-8 -*-

"""
MIT License

Copyright (c) 2019 Terbau

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import re
import datetime

class Store:
    """Object representing store data from Fortnite Battle Royale.
    
    Attributes
    ----------
    client: :class:`Client`
        The client.
    """
    def __init__(self, client, data):
        self.client = client
        self._daily_purchase_hours = data['dailyPurchaseHrs']
        self._refresh_interval_hours = data['refreshIntervalHrs']
        self._expires_at = self.client.from_iso(data['expiration'])

        self._featured_items = self._create_featured_items(data)
        self._daily_items = self._create_daily_items(data)

    @property
    def featured_items(self):
        """List[:class:`FeaturedStoreItem`]: A list containing data about
        featured items in the item shop.
        """
        return self._featured_items

    @property
    def daily_items(self):
        """List[:class:`DailyStoreItem`]: A list containing data about
        daily items in the item shop.
        """
        return self._daily_items

    @property
    def daily_purchase_hours(self):
        """:class:`int`: How many hours a day it is possible to purchase
        items. It most likely is ``24``.
        """
        return self._daily_purchase_hours

    @property
    def refresh_interval_hours(self):
        """:class:`int`: Refresh interval hours."""
        return self._refresh_interval_hours

    @property
    def created_at(self):
        """:class:`datetime.datetime`: The UTC time of the creation
        and current day.
        """
        return self._expires_at - datetime.timedelta(days=1)
    
    @property
    def expires_at(self):
        """:class:`datetime.datetime`: The UTC time of when this
        item shop expires.
        """
        return self._expires_at

    def _find_storefront(self, data, key):
        for storefront in data['storefronts']:
            if storefront['name'] == key:
                return storefront

    def _create_featured_items(self, data):
        storefront = self._find_storefront(data, 'BRWeeklyStorefront')

        res = []
        for item in storefront['catalogEntries']:
            res.append(FeaturedStoreItem(item))
        return res

    def _create_daily_items(self, data):
        storefront = self._find_storefront(data, 'BRDailyStorefront')

        res = []
        for item in storefront['catalogEntries']:
            res.append(DailyStoreItem(item))
        return res


class StoreItemBase:
    def __init__(self, data):
        self._dev_name = data['devName']
        self._asset_path = data.get('displayAssetPath')
   
        try:
            self._asset = re.search(r'\.(.+)', self._asset_path)[1]
        except TypeError:
            self._asset = None

        self._gifts_enabled = data['giftInfo']['bIsEnabled']
        self._daily_limit = data['dailyLimit']
        self._weekly_limit = data['weeklyLimit']
        self._monthly_limit = data['monthlyLimit']
        self._offer_id = data['offerId']
        self._offer_type = data['offerType']
        self._price = data['prices'][0]['finalPrice']
        self._refundable = data['refundable']
        self._items_grants = data['itemGrants']
        self._meta_info = data.get('metaInfo', [])
        self._meta = data.get('meta', {})

    @property
    def display_names(self):
        """List[:class:`str`]: The display names for this item."""
        match = re.search(r'^\[VIRTUAL][0-9]+ x (.*) for [0-9]+ .*$', self._dev_name)[1]
        return re.split(r', [0-9]+ x ', match)

    @property
    def dev_name(self):
        """:class:`str`: The dev name of this item."""
        return self._dev_name

    @property
    def asset_path(self):
        """:class:`str`: The asset path of the item. Could be ``None`` if
        not found.
        """
        return self._asset_path

    @property
    def asset(self):
        """:class:`str`: The asset of the item. Usually a CID or
        or something similar. Could be ``None`` if not found.
        """
        return self._asset

    @property
    def encryption_key(self):
        """:class:`str`: The encryption key for this item. If no encryption
        key is found, this will be ``None``.
        """
        for meta in self._meta_info:
            if meta['key'] == 'EncryptionKey':
                return meta['value']

    @property
    def gifts_enabled(self):
        """:class:`bool`: ``True`` if gifts is enabled for this
        item else ``False``.
        """
        return self._gifts_enabled

    @property
    def daily_limit(self):
        """:class:`int`: The daily account limit for this item.
        ``-1`` = Unlimited.
        """
        return self._daily_limit

    @property
    def weekly_limit(self):
        """:class:`int`: The weekly account limit for this item.
        ``-1`` = Unlimited.
        """
        return self._weekly_limit

    @property
    def monthly_limit(self):
        """:class:`int`: The monthly account limit for this item.
        ``-1`` = Unlimited.
        """
        return self._monthly_limit

    @property
    def offer_id(self):
        """:class:`str`: The offer id of this item."""
        return self._offer_id

    @property
    def offer_type(self):
        """:class:`str`: The offer type of this item."""
        return self._offer_type

    @property
    def price(self):
        """:class:`int`: The price of this item in v-bucks."""
        return self._price
    
    @property
    def refundable(self):
        """:class:`bool`: ``True`` if item is refundable else
        ``False``.
        """
        return self._refundable

    @property
    def grants(self):
        """:class:`list`: A list of items you get from this purchase.
        
        Typical output: ::
            
            [{
                'quantity': 1,
                'type': 'AthenaCharacter',
                'asset': 'cid_318_athena_commando_m_demon'
            }]  
        """
        l = []
        for item in self._items_grants:
            _type, _asset = item['templateId'].split(':')
            l.append({
                'quantity': item['quantity'],
                'type': _type,
                'asset': _asset
            })
        return l

    @property
    def new(self):
        """:class:`bool`: ``True`` if the item is in the item shop for
        the first time, else ``False``.
        """
        for meta in self._meta_info:
            if meta['value'].lower() == 'new':
                return True
        return False

    @property
    def violator(self):
        """:class:`str`: The violator of this item. Violator is the
        red tag at the top of an item in the shop. Will be ``None``
        if no violator is found for this item.
        """
        unfixed = self._meta.get('BannerOverride')
        if unfixed:
            return ' '.join(re.findall(r'[A-Z][^A-Z]*', unfixed))


class FeaturedStoreItem(StoreItemBase):
    """Featured store item."""
    def __init__(self, data):
        super().__init__(data)
        self._panel = int((data['categories'][0].split(' '))[1])

    @property
    def panel(self):
        """:class:`int`: The panel the item is listed in from left
        to right.
        """
        return self._panel


class DailyStoreItem(StoreItemBase):
    """Daily store item."""
    def __init__(self, data):
        super().__init__(data)



# -*- coding: utf-8 -*-

"""
MIT License

Copyright (c) 2019 Terbau

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import json
import re
import datetime

from .party import Party
from .errors import PartyPermissionError

class PresenceGameplayStats:
    """Represents gameplaystats received from presence.
    
    Attributes
    ----------
    state: :class:`str`
        The state.
    playlist: :class:`str`
        The playlist.
    num_kills: :class:`int`
        The amount of kills the friend currently has.
    fell_to_death: :class:`bool`
        ``True`` if friend fell to death in its current game, else ``False``    
    """

    __slots__ = ('state', 'playlist', 'num_kills', 'fell_to_death')

    def __init__(self, data):
        self.state = data.get('state')
        self.playlist = data.get('playlist')
        self.num_kills = data.get('numKills')
        if self.num_kills is not None:
            self.num_kills = int(self.num_kills)
        self.fell_to_death = data.get('bFellToDeath')
        if self.fell_to_death is not None:
            self.fell_to_death = bool(self.fell_to_death)


class PresenceParty:
    """Represents a party received from presence.

    Before accessing any of this class' attributes or functions
    you should always check if the party is private: ::
    
        @client.event
        async def event_friend_presence(presence):

            # check if presence is from the account 'Terbau'
            # NOTE: you should always use id over display_name
            # but for this example i've use display_name just
            # to demonstrate.
            if presence.author.display_name != 'Terbau':
                return
            
            # check if party is private
            if presence.party.is_private:
                return
            
            # if all the checks above succeeds we join the party
            await presence.party.join()
        
    
    .. note::

        If the party is private all attributes below is_private will
        be ``None``.

    Attributes
    ----------
    client: :class:`str`
        The client.
    is_private: :class:`bool`
        ``True`` if the party is private else ``False``.
    platform: :class:`str`
        The platform of the friend.
    id: :class:`str`
        The party's id.
    party_type_id: :class:`str`
        The party's type id.
    key: :class:`str`
        The party's key.
    app_id: :class:`str`
        The party's app id.
    build_id: :class:`str`
        The party's build id.
    party_flags: :class:`str`
        The party's flags.
    not_accepting_reason: :class:`str`
        The party's not accepting reason.
    playercount: :class:`int`
        The party's playercount.
    """

    __slots__ = ('client', 'is_private', 'platform', 'id', 'party_type_id',
                 'key', 'app_id', 'build_id', 'party_flags', 'not_accepting_reason',
                 'playercount')

    def __init__(self, client, data):
        self.client = client
        self.is_private = data.get('bIsPrivate', False)

        self.platform = data.get('sourcePlatform')
        self.id = data.get('partyId')
        self.party_type_id = data.get('partyTypeId')
        self.key = data.get('key')
        self.app_id = data.get('appId')
        self.build_id = data.get('buildId')
        self.party_flags = data.get('partyFlags')
        self.not_accepting_reason = data.get('notAcceptingReason')
        self.playercount = data.get('pc')
        if self.playercount is not None:
            self.playercount = int(self.playercount)

    async def join(self):
        """|coro|
        
        Joins the friends' party.

        Raises
        ------
        PartyPermissionError
            The party is private.
        HTTPException
            Something else went wrong when trying to join this party.
        """
        if self.is_private:
            raise PartyPermissionError('You cannot join a private party.')
        await self.client.join_to_party(self.id)


class Presence:
    """Represents a presence received from a friend
    
    Attributes
    ----------
    client: :class:`Client`
        The client.
    is_available: :class:`bool`
        The availability of this presence. ``True`` if presence is available,
        ``False`` if user went unavailable.
    friend: :class:`Friend`
        The friend you received this presence from.
    received_at: :class:`datetime.datetime`
        The UTC time of when the client received this presence.
    status: :class:`str`
        The friend's status.
    is_playing: :class:`bool`
        Says if friend is playing.
    is_joinable: :class:`bool`
        Says if friend is joinable.
    session_id: :class:`str`
        The friend's current session id. Often referred to as
        server key or game key.
    has_properties: :class:`bool`
        ``True`` if the presence has properties else ``False``.
        
        .. warning::

            All attributes below this point will be ``None`` if
            :attr:`has_properties` is ``False``.
    party: :class:`PresenceParty`
        The friend's party.
    gameplay_stats: :class:`PresenceGameplayStats`
        The friend's gameplay stats.
    homebase_rating: :class:`str`
        The friend's homebase rating
    lfg: :class:`bool`
        ``True`` if the friend is currently looking for a game.
    sub_game: :class:`str`
        The friend's current subgame.
    in_unjoinable_match: :class:`bool`
        ``True`` if friend is in unjoinable match else ``False``.
    playlist: :class:`str`
        The friend's current playlist.
    players_alive: :class:`int`
        The amount of players alive in the friend's current game.
    party_size: :class:`int`
        The size of the friend's party.
    max_party_size: :class:`int`
        The max size of the friend's party.
    game_session_join_key: :class:`str`
        The join key of the friend's session.
    server_player_count: :class:`str`
        The playercount of the friend's server.
    """

    __slots__ = ('client', 'is_available', 'friend', 'received_at', 'status', 'is_playing',
                 'is_joinable', 'has_voice_support', 'session_id', 'raw_properties',
                 'has_properties', 'homebase_rating', 'lfg', 'sub_game',
                 'in_unjoinable_match', 'playlist', 'players_alive', 'party_size',
                 'max_party_size', 'game_session_join_key', 'server_player_count',
                 'gameplay_stats', 'party')

    def __init__(self, client, from_id, is_available, data):
        self.client = client
        self.is_available = is_available
        self.friend = self.client.get_friend(from_id)
        self.received_at = datetime.datetime.utcnow()

        self.status = data['Status']
        self.is_playing = bool(data['bIsPlaying'])
        self.is_joinable = bool(data['bIsJoinable'])
        self.has_voice_support = bool(data['bHasVoiceSupport'])
        self.session_id = data['SessionId'] if data['SessionId'] != "" else None

        self.raw_properties = data['Properties']
        self.has_properties = self.raw_properties != {}
        
        # all values below will be "None" if properties is empty
        _basic_info = self.raw_properties.get('FortBasicInfo_j', {})
        self.homebase_rating = _basic_info.get('homeBaseRating')
        if self.raw_properties.get('FortLFG_I') is None:
            self.lfg = None
        else:
            self.lfg = True if int(self.raw_properties.get('FortLFG_I')) == 1 else False

        self.sub_game = self.raw_properties.get('FortSubGame_i')
        self.in_unjoinable_match = self.raw_properties.get('InUnjoinableMatch_b')
        if self.in_unjoinable_match is not None:
            self.in_unjoinable_match = int(self.in_unjoinable_match)
        self.playlist = self.raw_properties.get('GamePlaylistName_s')
        self.players_alive = self.raw_properties.get('Event_PlayersAlive_s')
        if self.players_alive is not None:
            self.players_alive = int(self.players_alive)
        self.party_size = self.raw_properties.get('Event_PartySize_s')
        if self.party_size is not None:
            self.party_size = int(self.party_size)
        self.max_party_size = self.raw_properties.get('Event_PartyMaxSize_s')
        if self.max_party_size is not None:
            self.max_party_size = int(self.max_party_size)
        self.game_session_join_key = self.raw_properties.get('GameSessionJoinKey_s')
        self.server_player_count = self.raw_properties.get('ServerPlayerCount_i')
        if self.server_player_count is not None:
            self.server_player_count = int(self.server_player_count)

        if 'FortGameplayStats_j' in self.raw_properties.keys():
            self.gameplay_stats = PresenceGameplayStats(self.raw_properties['FortGameplayStats_j'])
        else:
            self.gameplay_stats = None
        
        key = None
        for k in self.raw_properties.keys():
            if re.search(r'party\.joininfodata\.\d+_j', k) is not None:
                key = k
        if key is None:
            self.party = None
        else:
            self.party = PresenceParty(self.client, self.raw_properties[key])
